# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

r"""
Updated nucleic acid analysis --- :mod:`MDAnalysis.analysis.nucleicacids`
=========================================================================

:Author: Alia Lescoulie
:Year: 2022-2023
:copyright: LGPLv2.1

The module provides classes for analyzing nucleic acids structures.
This is an updated, higher performance version of previous nucleic acid tools.
For applications see :footcite:p:`Denning2011,Denning2012`.

.. rubric:: References

.. footbibliography::

Distances
---------

.. autoclass:: NucPairDist
    :members:
    :inherited-members:

.. autoclass:: WatsonCrickDist
    :members:
    :exclude-members: select_strand_atoms
    :inherited-members:

.. autoclass:: MinorPairDist
    :members:
    :exclude-members: select_strand_atoms
    :inherited-members:

.. autoclass:: MajorPairDist
    :members:
    :exclude-members: select_strand_atoms
    :inherited-members:

.. versionadded 2.2.0

"""

from typing import List, Dict, Tuple, Union
import warnings

import numpy as np

import MDAnalysis as mda
from .distances import calc_bonds
from .base import AnalysisBase, ResultsGroup
from MDAnalysis.core.groups import Residue, ResidueGroup


# Deprecation: In 3.0.0 change type to just
# ResidueClass = ResidueGroup
ResidueClass = Union[List[Residue], ResidueGroup]
r"""A type alias for :code:`Union[List[Residue], ResidueGroup]`

Used as an alias for methods where either class is acceptable.
"""


class NucPairDist(AnalysisBase):
    r"""Atom pair distance calculation base class.

    Takes two lists of :class:`~MDAnalysis.core.groups.AtomGroup` and
    computes the distances between them over a trajectory. Used as a
    superclass for the other nucleic acid distances classes. The distance
    will be measured between atoms sharing an index in the two lists of
    :class:`~MDAnalysis.core.groups.AtomGroup`.

    Parameters
    ----------
    selection1: List[AtomGroup]
        List of :class:`~MDAnalysis.core.groups.AtomGroup` containing an atom
        of each nucleic acid being analyzed.
    selection2: List[AtomGroup]
        List of :class:`~MDAnalysis.core.groups.AtomGroup` containing an atom
        of each nucleic acid being analyzed.
    kwargs: dict
        Arguments for :class:`~MDAnalysis.analysis.base.AnalysisBase`


    Attributes
    ----------
    results.pair_distances: numpy.ndarray
        2D array of pair distances. First dimension is simulation time,
        second dimension contains the pair distances for each each entry
        pair in selection1 and selection2.

        .. versionadded:: 2.4.0

        .. note::
            `results.pair_distances` is slated for deprecation in
            version 3.0.0, use `results.distances` instead.
        .. deprecated:: 2.7.0
            `results.pair_distances` will be removed in
            version 3.0.0, use :attr:`results.distances` instead.

    results.distances: numpy.ndarray
        stored in a 2d numpy array with first index selecting the
        Residue pair, and the second index selecting the frame number
        Distances are stored in a 2d numpy array with axis 0 (first index)
        indexing the trajectory frame and axis 1 (second index) selecting the
        Residue pair.

        .. versionadded:: 2.7.0

    times: numpy.ndarray
        Simulation times for analysis.


    Raises
    ------
    ValueError
        If the selections given are not the same length
    ValueError
        An :class:`~MDAnalysis.core.groups.AtomGroup` in one of the
        strands not a valid nucleic acid
    ValueError
        If a given residue pair from the provided strands returns an empty
        :class:`~MDAnalysis.core.groups.AtomGroup` when selecting the atom
        pairs used in the distance calculations


    *Version Info*

    .. versionchanged:: 2.5.0
       The ability to access by passing selection indices to :attr:`results`
       is now removed as of MDAnalysis version 2.5.0. Please use
       :attr:`results.pair_distances` instead.
       The :attr:`results.times` was deprecated and is now removed as of
       MDAnalysis 2.5.0.
       Please use the class attribute :attr:`times` instead.

    .. versionchanged:: 2.7.0
        Added static method :attr:`select_strand_atoms` as a
        helper for selecting atom pairs for distance analysis.

    .. versionchanged:: 2.9.0
       Enabled **parallel execution** with the ``multiprocessing`` and ``dask``
       backends; use the new method :meth:`get_supported_backends` to see all
       supported backends.
    """

    _analysis_algorithm_is_parallelizable = True

    @classmethod
    def get_supported_backends(cls):
        return ("serial", "multiprocessing", "dask")

    _s1: mda.AtomGroup
    _s2: mda.AtomGroup
    _n_sel: int

    def __init__(
        self,
        selection1: List[mda.AtomGroup],
        selection2: List[mda.AtomGroup],
        **kwargs,
    ) -> None:
        super(NucPairDist, self).__init__(
            selection1[0].universe.trajectory, **kwargs
        )

        if len(selection1) != len(selection2):
            raise ValueError("Selections must be same length")

        self._n_sel: int = len(selection1)

        self._s1 = selection1[0]
        self._s2 = selection2[0]

        for i in range(1, self._n_sel):
            self._s1 += selection1[i]
            self._s2 += selection2[i]

    @staticmethod
    def select_strand_atoms(
        strand1: ResidueGroup,
        strand2: ResidueGroup,
        a1_name: str,
        a2_name: str,
        g_name: str = "G",
        a_name: str = "A",
        u_name: str = "U",
        t_name: str = "T",
        c_name: str = "C",
    ) -> Tuple[List[mda.AtomGroup], List[mda.AtomGroup]]:
        r"""
        A helper method for nucleic acid pair distance analyses.
        Used for selecting specific atoms from two strands of nucleic acids.


        Parameters
        ----------
        strand1: List[Residue]
            The first nucleic acid strand
        strand2: List[Residue]
            The second nucleic acid strand
        a1_name: str
            The selection for the purine base of the strand pair
        a2_name: str
            the selection for the pyrimidine base of the strand pair
        g_name: str (optional)
            Name of Guanine in topology, by default assigned to G
        a_name: str (optional)
            Name of Adenine in topology, by default assigned to A
        u_name: str (optional)
            Name of Uracil in topology, by default assigned to U
        t_name: str (optional)
            Name of Thymine in topology, by default assigned to T
        c_name: str (optional)
            Name of Cytosine in topology, by default assigned to C

        Returns
        -------
        Tuple[List[AtomGroup], List[AtomGroup]]
            returns a tuple containing two lists of
            :class:`~MDAnalysis.core.groups.AtomGroup`\s
            corresponding to the provided selections from each strand.

        Raises
        ------
        ValueError:
            An :class:`~MDAnalysis.core.groups.AtomGroup`
            in one of the strands not a valid nucleic acid
        ValueError:
            An :class:`~MDAnalysis.core.groups.Residue` returns an empty
            :class:`~MDAnalysis.core.groups.AtomGroup`
            with the provided selection


        .. versionadded:: 2.7.0
        """
        pyrimidines: List[str] = [c_name, t_name, u_name]
        purines: List[str] = [a_name, g_name]

        sel1: List[mda.AtomGroup] = []
        sel2: List[mda.AtomGroup] = []

        for pair in zip(strand1.residues, strand2.residues):
            if pair[0].resname[0] in pyrimidines:
                a1, a2 = a2_name, a1_name
            elif pair[0].resname[0] in purines:
                a1, a2 = a1_name, a2_name
            else:
                raise ValueError(
                    f"AtomGroup in {pair} is not a valid nucleic acid"
                )

            ag1 = pair[0].atoms.select_atoms(f"name {a1}")
            ag2 = pair[1].atoms.select_atoms(f"name {a2}")

            if not all(len(ag) > 0 for ag in [ag1, ag2]):
                err_info: Tuple[Residue, str] = (
                    (pair[0], a1) if len(ag1) == 0 else (pair[1], a2)
                )

                raise ValueError(
                    (
                        f"{err_info[0]} returns an empty AtomGroup"
                        'with selection string "name {a2}"'
                    )
                )

            sel1.append(ag1)
            sel2.append(ag2)

        return (sel1, sel2)

    def _prepare(self) -> None:
        self.results.distances: np.ndarray = np.zeros(
            [self.n_frames, self._n_sel]
        )

    def _single_frame(self) -> None:
        dist: np.ndarray = calc_bonds(self._s1.positions, self._s2.positions)

        self.results.distances[self._frame_index, :] = dist

    def _conclude(self) -> None:
        self.results["pair_distances"] = self.results["distances"]
        # TODO: remove pair_distances in 3.0.0

    def _get_aggregator(self):
        return ResultsGroup(
            lookup={
                "distances": ResultsGroup.ndarray_vstack,
            }
        )


class WatsonCrickDist(NucPairDist):
    r"""
    Watson-Crick base pair distance for selected
    residues over a trajectory.

    Takes two :class:`~MDAnalysis.core.groups.ResidueGroup`
    objects or two lists of :class:`~MDAnalysis.core.groups.Residue`
    and calculates the distance between the nitrogen atoms in the
    Watson-Crick hydrogen bond over the trajectory. Bases are matched
    either by their index in the two
    :class:`~MDAnalysis.core.groups.ResidueGroup` provided as arguments,
    or based on the indices of the provided lists of
    :class:`~MDAnalysis.core.groups.Residue` objects depending
    on which is provided.

    .. note::
        Support for :class:`~MDAnalysis.core.groups.Residue` is slated for
        deprecation and will raise a warning when used. It still works but
        :class:`~MDAnalysis.core.groups.ResidueGroup` is preferred.

    Parameters
    ----------
    strand1: ResidueClass
        First list of bases

        .. deprecated:: 2.7.0
           Using a list of :class:`~MDAnalysis.core.groups.Residue` will
           be removed in 3.0.0. Use a
           :class:`~MDAnalysis.core.groups.ResidueGroup`.

    strand2: ResidueClass
        Second list of bases

        .. deprecated:: 2.7.0
           Using a list of :class:`~MDAnalysis.core.groups.Residue` will
           be removed in 3.0.0. Use a
           :class:`~MDAnalysis.core.groups.ResidueGroup`.

    n1_name: str (optional)
        Name of Nitrogen 1 of nucleic acids, by default assigned to "N1"
    n3_name: str (optional)
        Name of Nitrogen 3 of nucleic acids, by default assigned to "N3"
    g_name: str (optional)
        Name of Guanine in topology, by default assigned to "G"
    a_name: str (optional)
        Name of Adenine in topology, by default assigned to "A"
    u_name: str (optional)
        Name of Uracil in topology, by default assigned to "U"
    t_name: str (optional)
        Name of Thymine in topology, by default assigned to "T"
    c_name: str (optional)
        Name of Cytosine in topology, by default assigned to C
    **kwargs: dict
        Key word arguments for
        :class:`~MDAnalysis.analysis.base.AnalysisBase`

    Attributes
    ----------
    results.distances: numpy.ndarray
        Distances are stored in a 2d numpy array with axis 0 (first index)
        indexing the trajectory frame and axis 1 (second index) selecting the
        Residue pair.

        .. versionadded:: 2.7.0

    results.pair_distances: numpy.ndarray
        2D array of pair distances. First dimension is
        simulation time, second dimension contains the
        pair distances for each each entry pair in
        selection1 and selection2.

        .. versionadded:: 2.4.0

        .. deprecated:: 2.7.0
            `results.pair_distances` will be removed in version 3.0.0,
            use :attr:`results.distances` instead.

    times: numpy.ndarray
        Simulation times for analysis.

    Raises
    ------
    TypeError
        If the provided list of :class:`~MDAnalysis.core.Residue` contains
        non-Residue elements

        .. deprecated:: 2.7.0
           Starting with version 3.0.0, this exception will no longer
           be raised because only
           :class:`~MDAnalysis.core.groups.ResidueGroup` will be allowed.

    ValueError
        If `strand1` and `strand2` are not the same length
    ValueError:
        An :class:`~MDAnalysis.core.groups.AtomGroup`
        in one of the strands not a valid nucleic acid
    ValueError
        If a given residue pair from the provided strands returns an empty
        :class:`~MDAnalysis.core.groups.AtomGroup` when selecting the atom
        pairs used in the distance calculations


    *Version Info*

    .. versionchanged:: 2.5.0
       Accessing results by passing strand indices to :attr:`results`
       was deprecated and is now removed as of MDAnalysis version 2.5.0.
       Please use :attr:`results.pair_distances` instead.
       The :attr:`results.times` was deprecated and is now removed as of
       MDAnalysis 2.5.0. Please use the class attribute
       :attr:`times` instead.

    .. versionchanged:: 2.7.0
        `strand1` and `strand2` now also accept a
        :class:`~MDAnalysis.core.groups.ResidueGroup` as input.
        The previous input type, ``List[Residue]`` is still supported,
        but it is **deprecated** and will be removed in release 3.0.0.
    """

    def __init__(
        self,
        strand1: ResidueClass,
        strand2: ResidueClass,
        n1_name: str = "N1",
        n3_name: str = "N3",
        g_name: str = "G",
        a_name: str = "A",
        u_name: str = "U",
        t_name: str = "T",
        c_name: str = "C",
        **kwargs,
    ) -> None:

        def verify_strand(strand: ResidueClass) -> ResidueGroup:
            # Helper method to verify the strands

            if isinstance(strand, list):  # Checking if a list is given
                # verify list is only Residues
                if not all(isinstance(resid, Residue) for resid in strand):
                    raise TypeError(f"{strand} contains non-Residue elements")

                warnings.warn(
                    DeprecationWarning(
                        (
                            f"ResidueGroup should be used for {strand} instead"
                            "of giving a Residue list"
                        )
                    )
                )
                # Convert to a ResidueGroup
                strand: ResidueGroup = ResidueGroup(strand)

            return strand

        strand1: ResidueGroup = verify_strand(strand1)
        strand2: ResidueGroup = verify_strand(strand2)

        strand_atomgroups: Tuple[List[mda.AtomGroup], List[mda.AtomGroup]] = (
            self.select_strand_atoms(
                strand1,
                strand2,
                n1_name,
                n3_name,
                g_name=g_name,
                a_name=a_name,
                t_name=t_name,
                u_name=u_name,
                c_name=c_name,
            )
        )

        super(WatsonCrickDist, self).__init__(
            strand_atomgroups[0], strand_atomgroups[1], **kwargs
        )


class MinorPairDist(NucPairDist):
    r"""Minor-Pair basepair distance for selected residues over a trajectory.

    Takes two :class:`~MDAnalysis.core.groups.ResidueGroup` objects and
    calculates the Minor-groove hydrogen bond length between the
    nitrogen and oxygen atoms over the trajectory. Bases are
    matched by their index in the two
    :class:`~MDAnalysis.core.groups.ResidueGroup` provided as arguments.

    Parameters
    ----------
    strand1: List[Residue]
        First list of bases
    strand2: List[Residue]
        Second list of bases
    o2_name: str (optional)
        Name of Oxygen 2 of nucleic acids;
        by default assigned to "O2";
    c2_name: str (optional)
        Name of Carbon 2 of nucleic acids;
        by default assigned to "C2";
    g_name: str (optional)
        Name of Guanine in topology;
        by default assigned to "G";
    a_name: str (optional)
        Name of Adenine in topology
        by default assigned to "A";
    u_name: str (optional)
        Name of Uracil in topology;
        by default assigned to "U";
    t_name: str (optional)
        Name of Thymine in topology;
        by default assigned to "T";
    c_name: str (optional)
        Name of Cytosine in topology;
        by default assigned to "C";
    **kwargs:
        keyword arguments for
        :class:`~MDAnalysis.analysis.base.AnalysisBase`

    Attributes
    ----------
    results.distances: numpy.ndarray
        stored in a 2d numpy array with first index selecting
        the Residue pair, and the second index selecting the frame number
    times: numpy.ndarray
        Simulation times for analysis.

    Raises
    ------
    ValueError
        If the selections given are not the same length
        A :class:`~MDAnalysis.core.Residue` in
        one of the strands not a valid nucleic acid
    ValueError
        If a given residue pair from the provided strands returns an empty
        :class:`~MDAnalysis.core.groups.AtomGroup` when selecting the atom
        pairs used in the distance calculations


    .. versionadded:: 2.7.0
    """

    def __init__(
        self,
        strand1: ResidueGroup,
        strand2: ResidueGroup,
        o2_name: str = "O2",
        c2_name: str = "C2",
        g_name: str = "G",
        a_name: str = "A",
        u_name: str = "U",
        t_name: str = "T",
        c_name: str = "C",
        **kwargs,
    ) -> None:

        selections: Tuple[List[mda.AtomGroup], List[mda.AtomGroup]] = (
            self.select_strand_atoms(
                strand1,
                strand2,
                c2_name,
                o2_name,
                g_name=g_name,
                a_name=a_name,
                t_name=t_name,
                u_name=u_name,
                c_name=c_name,
            )
        )

        super(MinorPairDist, self).__init__(
            selections[0], selections[1], **kwargs
        )


class MajorPairDist(NucPairDist):
    r"""Minor-Pair base pair distance for
    selected residues over a trajectory.

    Takes two :class:`~MDAnalysis.core.groups.ResidueGroup` objects and
    calculates the Major-groove hydrogen bond length between the nitrogen
    and oxygen atoms over the trajectory. Bases are matched by their index
    in the two :class:`~MDAnalysis.core.groups.ResidueGroup`
    provided as arguments.

    Parameters
    ----------
    strand1: List[Residue]
        First list of bases
    strand2: List[Residue]
        Second list of bases
    o6_name: str (optional)
        Name of Oxygen 6 of nucleic acids;
        by default assigned to "O6"
    n4_name: str (optional)
        Name of Nitrogen 4 of nucleic acids;
        by default assigned to "N4"
    g_name: str (optional)
        Name of Guanine in topology;
        by default assigned to "G"
    a_name: str (optional)
        Name of Adenine in topology;
        by default assigned to "A"
    u_name: str (optional)
        Name of Uracil in topology;
        by default assigned to "U"
    t_name: str (optional)
        Name of Thymine in topology;
        by default assigned to "T"
    c_name: str (optional)
        Name of Cytosine in topology;
        by default assigned to "C"
    **kwargs:
        arguments for :class:`~MDAnalysis.analysis.base.AnalysisBase`

    Attributes
    ----------
    results.distances: numpy.ndarray
        Distances are stored in a 2d numpy array with axis 0 (first index)
        indexing the trajectory frame and axis 1 (second index) selecting the
        Residue pair.
    times: numpy.ndarray
        Simulation times for analysis.

    Raises
    ------
    ValueError
        A :class:`~MDAnalysis.core.Residue`
        in one of the strands not a valid nucleic acid
    ValueError
        If a given residue pair from the provided strands returns an empty
        :class:`~MDAnalysis.core.groups.AtomGroup` when selecting the atom
        pairs used in the distance calculations
    ValueError
        if the selections given are not the same length


    .. versionadded:: 2.7.0
    """

    def __init__(
        self,
        strand1: ResidueGroup,
        strand2: ResidueGroup,
        n4_name: str = "N4",
        o6_name: str = "O6",
        g_name: str = "G",
        a_name: str = "A",
        u_name: str = "U",
        t_name: str = "T",
        c_name: str = "C",
        **kwargs,
    ) -> None:

        selections: Tuple[List[mda.AtomGroup], List[mda.AtomGroup]] = (
            self.select_strand_atoms(
                strand1,
                strand2,
                o6_name,
                n4_name,
                g_name=g_name,
                a_name=a_name,
                t_name=t_name,
                u_name=u_name,
                c_name=c_name,
            )
        )

        super(MajorPairDist, self).__init__(
            selections[0], selections[1], **kwargs
        )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
TPR file format --- :mod:`MDAnalysis.coordinates.TPR`
======================================================

Class for reading positions and velocities from GROMACS TPR files.


Reading TPR files
-----------------
MDAnalysis can read positions and velocities from GROMACS TPR files,
and can also do so using different versions of the topology and
coordinate file for a given system.


For example, both of these are supported modes for reading
in positions and velocities from GROMACS TPR files::

   >>> u = mda.Universe(TPR2020, TPR2024_4)
   >>> u = mda.Universe(TPR2020)


Classes
-------

.. autoclass:: TPRReader
   :members:
   :inherited-members:

"""

from . import base
from ..lib import util
from .timestep import Timestep
import MDAnalysis.topology.tpr.utils as tpr_utils
import MDAnalysis.topology.tpr.setting as S

import logging

logger = logging.getLogger("MDAnalysis.coordinates.TPR")

import numpy as np


class TPRReader(base.SingleFrameReaderBase):
    # TODO: reduce duplication with `TPRparser`;
    # we could share some state for the position
    # in the binary file to avoid re-reading topology
    # or perhaps combine the topology and coordinate reading
    # with some inheritance shenanigans?
    """Class supporting read in of positions and velocities from GROMACS TPR files.

    .. versionadded:: 2.10.0
    """
    format = "TPR"
    units = {"length": "nm", "velocity": "nm/ps"}
    _Timestep = Timestep

    def _read_first_frame(self):
        # Read header/move over topology
        # TODO: reduce duplication with TPRparser perhaps...
        with util.openany(self.filename, mode="rb") as infile:
            tprf = infile.read()
        data = tpr_utils.TPXUnpacker(tprf)
        try:
            th = tpr_utils.read_tpxheader(data)  # tpxheader
        except (EOFError, ValueError):
            msg = f"{self.filename}: Invalid tpr coordinate file or cannot be recognized"
            logger.critical(msg)
            raise IOError(msg)

        self.ts = ts = self._Timestep(th.natoms, **self._ts_kwargs)
        self.n_atoms = th.natoms

        # Starting with gromacs 2020 (tpx version 119), the body of the file
        # is encoded differently. We change the unpacker accordingly.
        if th.fver >= S.tpxv_AddSizeField and th.fgen >= 27:
            actual_body_size = len(data.get_buffer()) - data.get_position()
            if actual_body_size == 4 * th.sizeOfTprBody:
                # See issue #2428.
                msg = (
                    "TPR files produced with beta versions of gromacs 2020 "
                    "are not supported."
                )
                logger.critical(msg)
                raise IOError(msg)
            data = tpr_utils.TPXUnpacker2020.from_unpacker(data)

        state_ngtc = th.ngtc  # done init_state() in src/gmxlib/tpxio.c
        if th.bBox:
            tpr_utils.extract_box_info(data, th.fver)

        if state_ngtc > 0:
            if th.fver < 69:  # redundancy due to  different versions
                tpr_utils.ndo_real(data, state_ngtc)
            tpr_utils.ndo_real(
                data, state_ngtc
            )  # relevant to Berendsen tcoupl_lambda

        tpr_top = tpr_utils.do_mtop(
            data, th.fver, tpr_resid_from_one=True, precision=th.precision
        )

        if th.bX:
            self.ts._pos = np.asarray(
                tpr_utils.ndo_rvec(data, th.natoms), dtype=np.float32
            )
        if th.bV:
            self.ts.velocities = np.asarray(
                tpr_utils.ndo_rvec(data, th.natoms), dtype=np.float32
            )
            self.ts.has_velocities = True



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

""" Base classes --- :mod:`MDAnalysis.converters.base`
======================================================

Converters output information to other libraries.

.. autoclass:: ConverterBase
   :members:
   :inherited-members:
"""

from .. import _CONVERTERS
from ..coordinates.base import IOBase
from ..lib.util import asiterable


class _Convertermeta(type):
    # Auto register upon class creation
    def __init__(cls, name, bases, classdict):
        type.__init__(type, name, bases, classdict)
        try:
            fmt = asiterable(classdict["lib"])
        except KeyError:
            pass
        else:
            for f in fmt:
                f = f.upper()
                _CONVERTERS[f] = cls


class ConverterBase(IOBase, metaclass=_Convertermeta):
    """Base class for converting to other libraries."""

    def __repr__(self):
        return "<{cls}>".format(cls=self.__class__.__name__)

    def convert(self, obj):
        raise NotImplementedError



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
PDB Topology Parser
=========================================================================

This topology parser uses a standard PDB file to build a minimum
internal structure representation (list of atoms).

The topology reader reads a PDB file line by line and ignores atom
numbers but only reads residue numbers up to 9,999 correctly. If you
have systems containing at least 10,000 residues then you need to use
a different file format (e.g. the "extended" PDB, *XPDB* format, see
:mod:`~MDAnalysis.topology.ExtendedPDBParser`) that can handle residue
numbers up to 99,999.

.. note::

    Atomtypes will be created from elements if they are present and valid.
    Otherwise, they will be guessed on Universe creation.
    By default, masses will also be guessed on Universe creation.
    This may change in release 3.0.
    See :ref:`Guessers` for more information.


.. Note::

   The parser processes atoms and their names.
   Partial charges are not set. Elements are parsed if they are
   valid. If partially missing or incorrect, empty records are assigned.

See Also
--------
* :mod:`MDAnalysis.topology.ExtendedPDBParser`
* :class:`MDAnalysis.coordinates.PDB.PDBReader`
* :class:`MDAnalysis.core.universe.Universe`


Classes
-------

.. autoclass:: PDBParser
   :members:
   :inherited-members:

"""
import numpy as np
import warnings
import logging

from ..guesser.tables import SYMB2Z
from ..lib import util
from .base import TopologyReaderBase, change_squash
from ..core.topology import Topology
from ..core.topologyattrs import (
    Atomnames,
    Atomids,
    AltLocs,
    Bonds,
    ChainIDs,
    Atomtypes,
    Elements,
    ICodes,
    Occupancies,
    RecordTypes,
    Resids,
    Resnames,
    Resnums,
    Segids,
    Tempfactors,
    FormalCharges,
)

# Set up a logger for the PDBParser
logger = logging.getLogger("MDAnalysis.topology.PDBParser")


def float_or_default(val, default):
    try:
        return float(val)
    except ValueError:
        return default


DIGITS_UPPER = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
DIGITS_LOWER = DIGITS_UPPER.lower()
DIGITS_UPPER_VALUES = dict([pair for pair in zip(DIGITS_UPPER, range(36))])
DIGITS_LOWER_VALUES = dict([pair for pair in zip(DIGITS_LOWER, range(36))])


def decode_pure(digits_values, s):
    """Decodes the string s using the digit, value associations for each
    character.

    Parameters
    ----------
    digits_values: dict
        A dictionary containing the base-10 numbers that each hexadecimal
        number corresponds to.
    s: str
        The contents of the pdb index columns.

    Returns
    -------
    The integer in base-10 corresponding to traditional base-36.
    """
    result = 0
    n = len(digits_values)
    for c in s:
        result *= n
        result += digits_values[c]
    return result


def hy36decode(width, s):
    """
    Decodes base-10/upper-case base-36/lower-case base-36 hybrid.

    Parameters
    ----------
    width: int
        The number of columns in the pdb file store atom index.
    s: str
        The contents of the pdb index columns.

    Returns
    -------
    int
        Base-10 integer corresponding to hybrid36.
    """
    if (len(s) == width):
        f = s[0]
        if (f == "-" or f == " " or f.isdigit()):
            return int(s)
        elif (f in DIGITS_UPPER_VALUES):
            return decode_pure(digits_values=DIGITS_UPPER_VALUES,
                               s=s) - 10 * 36 ** (width - 1) + 10 ** width
        elif (f in DIGITS_LOWER_VALUES):
            return decode_pure(digits_values=DIGITS_LOWER_VALUES,
                               s=s) + 16 * 36 ** (width - 1) + 10 ** width
    raise ValueError("invalid number literal.")


class PDBParser(TopologyReaderBase):
    """Parser that obtains a list of atoms from a standard PDB file.

    Creates the following Attributes (if present):
     - names
     - chainids
     - tempfactors
     - occupancies
     - record_types (ATOM/HETATM)
     - resids
     - resnames
     - segids
     - elements
     - bonds
     - formalcharges

    Note that `PDBParser` accepts an optional keyword argument
    ``force_chainids_to_segids``. If set to ``True``, the chain IDs (even if
    empty values are in the chain ID column in the file) will forcibly be used
    instead of the segment IDs for creating segments.

    See Also
    --------
    :class:`MDAnalysis.coordinates.PDB.PDBReader`


    .. versionadded:: 0.8
    .. versionchanged:: 0.18.0
       Added parsing of Record types
    .. versionchanged:: 1.0.0
       Added parsing of valid Elements
    .. versionchanged:: 2.0.0
       Bonds attribute is not added if no bonds are present in PDB file.
       If elements are invalid or partially missing, empty elements records
       are now assigned (Issue #2422).
       Aliased ``bfactors`` topologyattribute to ``tempfactors``.
       ``bfactors`` is deprecated and will be removed in 3.0 (Issue #1901)
    .. versionchanged:: 2.3.0
       Formal charges are now read from PDB files if present. No formalcharge
       attribute is created if no formal charges are present in the PDB file.
       Any formal charges not set are assumed to have a value of 0.
       Raise `UserWarning` instead `RuntimeError`
       when CONECT records are corrupt.
    .. versionchanged:: 2.5.0
       Formal charges will not be populated if an unknown entry is encountered,
       instead a UserWarning is emitted.
    .. versionchanged:: 2.8.0
        Removed type and mass guessing (attributes guessing takes place now
        through universe.guess_TopologyAttrs() API).
    .. versionchanged:: 2.10.0
        segID is read from 73-76 instead of 67-76 and added the
        `force_chainids_to_segids` keyword argument. Some infos in logger will
        be generated if the segids is not present or if the chainids are not
        completely equal to segids.
    """
    format = ['PDB', 'ENT']

    def parse(self, **kwargs):
        """Parse atom information from PDB file

        Returns
        -------
        MDAnalysis Topology object
        """
        top = self._parseatoms(**kwargs)

        try:
            bonds = self._parsebonds(top.ids.values)
        except AttributeError:
            warnings.warn("Invalid atom serials were present, "
                          "bonds will not be parsed")
        except RuntimeError:
            warnings.warn("CONECT records was corrupt, "
                          "bonds will not be parsed")
        else:
            # Issue 2832: don't append Bonds if there are no bonds
            if bonds:
                top.add_TopologyAttr(bonds)

        return top

    def _parseatoms(self, **kwargs):
        """Create the initial Topology object"""
        resid_prev = 0  # resid looping hack

        record_types = []
        serials = []
        names = []
        altlocs = []
        chainids = []
        icodes = []
        tempfactors = []
        occupancies = []
        resids = []
        resnames = []
        segids = []
        elements = []
        formalcharges = []

        self._wrapped_serials = False  # did serials go over 100k?
        last_wrapped_serial = 100000  # if serials wrap, start from here
        with util.openany(self.filename) as f:
            for line in f:
                line = line.strip()  # Remove extra spaces
                if not line:  # Skip line if empty
                    continue
                if line.startswith('END'):
                    break
                if not line.startswith(('ATOM', 'HETATM')):
                    continue

                record_types.append(line[:6].strip())
                try:
                    serial = int(line[6:11])
                except:
                    try:
                        serial = hy36decode(5, line[6:11])
                    except ValueError:
                        # serial can become '***' when they get too high
                        self._wrapped_serials = True
                        serial = last_wrapped_serial
                        last_wrapped_serial += 1
                finally:
                    serials.append(serial)

                names.append(line[12:16].strip())
                altlocs.append(line[16:17].strip())
                resnames.append(line[17:21].strip())
                chainids.append(line[21:22].strip())
                elements.append(line[76:78].strip())
                formalcharges.append(line[78:80].strip())

                # Resids are optional
                try:
                    if self.format == "XPDB":  # fugly but keeps code DRY
                        # extended non-standard format used by VMD
                        resid = int(line[22:27])
                    else:
                        resid = int(line[22:26])
                        # Wrapping
                        while resid - resid_prev < -5000:
                            resid += 10000
                        resid_prev = resid
                except ValueError:
                    warnings.warn("PDB file is missing resid information.  "
                                  "Defaulted to '1'")
                    resid = 1
                finally:
                    resids.append(resid)
                    icodes.append(line[26:27].strip())

                occupancies.append(float_or_default(line[54:60], 0.0))
                tempfactors.append(float_or_default(line[60:66], 1.0))  # AKA bfactor

                segids.append(line[72:76].strip())

        # Warn about wrapped serials
        if self._wrapped_serials:
            warnings.warn("Serial numbers went over 100,000.  "
                          "Higher serials have been guessed")

        # If segids is not equal to chainids, warn the user
        if any([a != b for a, b in zip(segids, chainids)]):
            logger.debug("Segment IDs and Chain IDs are not completely equal.")

        # If segids not present, try to use chainids
        if not any(segids):
            logger.info("Setting segids from chainIDs because no segids "
                        "found in the PDB file.")
            segids = chainids

        # If force_chainids_to_segids is set, use chainids as segids
        if kwargs.get("force_chainids_to_segids", False):
            logger.info("force_chainids_to_segids is set. "
                        "Using chain IDs as segment IDs.")
            segids = chainids

        n_atoms = len(serials)

        attrs = []
        # Make Atom TopologyAttrs
        for vals, Attr, dtype in (
                (names, Atomnames, object),
                (altlocs, AltLocs, object),
                (chainids, ChainIDs, object),
                (record_types, RecordTypes, object),
                (serials, Atomids, np.int32),
                (tempfactors, Tempfactors, np.float32),
                (occupancies, Occupancies, np.float32),
        ):
            attrs.append(Attr(np.array(vals, dtype=dtype)))
        # OPT: We do this check twice, maybe could refactor to avoid this
        if any(elements):
            # Feed atomtypes as raw element column, but validate elements
            atomtypes = elements
            attrs.append(Atomtypes(np.array(elements, dtype=object)))

            validated_elements = []
            for elem in elements:
                if elem.capitalize() in SYMB2Z:
                    validated_elements.append(elem.capitalize())
                else:
                    wmsg = (f"Unknown element {elem} found for some atoms. "
                            f"These have been given an empty element record. "
                            f"If needed they can be guessed using "
                            f"universe.guess_TopologyAttrs(context='default',"
                            " to_guess=['elements']).")
                    warnings.warn(wmsg)
                    validated_elements.append('')
            attrs.append(Elements(np.array(validated_elements, dtype=object)))
        else:
            warnings.warn("Element information is missing, elements attribute "
                          "will not be populated. If needed these can be"
                          " guessed using universe.guess_TopologyAttrs("
                          "context='default', to_guess=['elements']).")

        if any(formalcharges):
            try:
                for i, entry in enumerate(formalcharges):
                    if not entry == '':
                        if entry == '0':
                            # Technically a lack of charge shouldn't be in the
                            # PDB but MDA has a few files that specifically
                            # have 0 entries, indicating that some folks
                            # interpret 0 as an allowed entry
                            formalcharges[i] = 0
                        elif ('+' in entry) or ('-' in entry):
                            formalcharges[i] = int(entry[::-1])
                        else:
                            raise ValueError
                    else:
                        formalcharges[i] = 0
            except ValueError:
                wmsg = (f"Unknown entry {entry} encountered in formal charge "
                        "field. This likely indicates that the PDB file is "
                        "not fully standard compliant. The formalcharges "
                        "attribute will not be populated.")
                warnings.warn(wmsg)
            else:
                attrs.append(FormalCharges(np.array(formalcharges, dtype=int)))

        # Residue level stuff from here
        resids = np.array(resids, dtype=np.int32)
        resnames = np.array(resnames, dtype=object)
        if self.format == 'XPDB':  # XPDB doesn't have icodes
            icodes = [''] * n_atoms
        icodes = np.array(icodes, dtype=object)
        resnums = resids.copy()
        segids = np.array(segids, dtype=object)

        residx, (resids, resnames, icodes, resnums, segids) = change_squash(
            (resids, resnames, icodes, segids), (resids, resnames, icodes, resnums, segids))
        n_residues = len(resids)
        attrs.append(Resnums(resnums))
        attrs.append(Resids(resids))
        attrs.append(ICodes(icodes))
        attrs.append(Resnames(resnames))

        if (
            kwargs.get("force_chainids_to_segids", False) or
            (any(segids) and not any(val is None for val in segids))
        ):
            segidx, (segids,) = change_squash((segids,), (segids,))
            n_segments = len(segids)
            attrs.append(Segids(segids))
        else:
            n_segments = 1
            attrs.append(Segids(np.array(['SYSTEM'], dtype=object)))
            segidx = None
            logger.info("Segment/chain ID is empty, "
                        "setting segids to default value 'SYSTEM'.")

        top = Topology(n_atoms, n_residues, n_segments,
                       attrs=attrs,
                       atom_resindex=residx,
                       residue_segindex=segidx)

        return top

    def _parsebonds(self, serials):
        # Could optimise this by saving lines in the main loop
        # then doing post processing after all Atoms have been read
        # ie do one pass through the file only
        # Problem is that in multiframe PDB, the CONECT is at end of file,
        # so the "break" call happens before bonds are reached.

        # If the serials wrapped, this won't work
        if self._wrapped_serials:
            warnings.warn("Invalid atom serials were present, bonds will not"
                          " be parsed")
            raise AttributeError  # gets caught in parse

        # Mapping between the atom array indicies a.index and atom ids
        # (serial) in the original PDB file
        mapping = dict((s, i) for i, s in enumerate(serials))

        bonds = set()
        with util.openany(self.filename) as f:
            lines = (line for line in f if line[:6] == "CONECT")
            for line in lines:
                atom, atoms = _parse_conect(line.strip())
                for a in atoms:
                    try:
                        bond = tuple([mapping[atom], mapping[a]])
                    except KeyError:
                        # Bonds to TER records have no mapping
                        # Ignore these as they are not real atoms
                        warnings.warn(
                            "PDB file contained CONECT record to TER entry. "
                            "These are not included in bonds.")
                    else:
                        bonds.add(bond)

        bonds = tuple(bonds)

        return Bonds(bonds)


def _parse_conect(conect):
    """parse a CONECT record from pdbs

    Parameters
    ----------
    conect : str
        white space striped CONECT record

    Returns
    -------
    atom_id : int
        atom index of bond
    bonds : set
        atom ids of bonded atoms

    Raises
    ------
    RuntimeError
        Raised if ``conect`` is not a valid CONECT record
    """
    atom_id = int(conect[6:11])
    n_bond_atoms = len(conect[11:]) // 5

    try:
        if len(conect[11:]) % n_bond_atoms != 0:
            raise RuntimeError("Bond atoms aren't aligned proberly for CONECT "
                               "record: {}".format(conect))
    except ZeroDivisionError:
        # Conect record with only one entry (CONECT A\n)
        warnings.warn("Found CONECT record with single entry, ignoring this")
        return atom_id, []  # return empty list to allow iteration over nothing

    bond_atoms = (int(conect[11 + i * 5: 16 + i * 5]) for i in
                  range(n_bond_atoms))
    return atom_id, bond_atoms



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""\
=========================================================
Core object: Universe --- :mod:`MDAnalysis.core.universe`
=========================================================

The :class:`~MDAnalysis.core.universe.Universe` class ties a topology
and a trajectory together. Almost all code in MDAnalysis starts with a
``Universe``.

Normally, a ``Universe`` is created from files::

  import MDAnalysis as mda
  u = mda.Universe("topology.psf", "trajectory.dcd")

In order to construct new simulation system it is also convenient to
construct a ``Universe`` from existing
:class:`~MDAnalysis.core.group.AtomGroup` instances with the
:func:`Merge` function.


Classes
=======

.. autoclass:: Universe
   :members:

Functions
=========

.. autofunction:: Merge

"""
import errno
import numpy as np
import logging
import copy
import warnings
import contextlib
import collections

import MDAnalysis
import sys

# When used in an MPI environment with Infiniband, importing MDAnalysis may
# trigger an MPI warning because importing the uuid module triggers a call to
# os.fork(). This happens if MPI_Init() has been called prior to importing
# MDAnalysis. The problem is actually caused by the uuid module and not by
# MDAnalysis itself. Python 3.7 fixes the problem.
import os
import uuid

from .. import _TOPOLOGY_ATTRS, _PARSERS
from ..exceptions import NoDataError
from ..lib import util
from ..lib.log import ProgressBar
from ..lib.util import cached, NamedStream, isstream
from ..lib.mdamath import find_fragments
from . import groups
from ._get_readers import get_reader_for, get_parser_for
from .groups import (
    ComponentBase,
    GroupBase,
    Atom,
    Residue,
    Segment,
    AtomGroup,
    ResidueGroup,
    SegmentGroup,
)
from .topology import Topology
from .topologyattrs import (
    AtomAttr,
    ResidueAttr,
    SegmentAttr,
    Segindices,
    Segids,
    Resindices,
    Resids,
    Atomindices,
    BFACTOR_WARNING,
    _Connection,
)
from .topologyobjects import TopologyObject
from ..guesser.base import get_guesser

logger = logging.getLogger("MDAnalysis.core.universe")


def _update_topology_by_ids(universe, atomwise_resids, atomwise_segids):
    """Update the topology of a Universe with new atomwise resids and segids.

    Parameters
    ----------
    universe : Universe
        The universe to update.
    atomwise_resids : numpy.ndarray
        The new atomwise residue indices.
    atomwise_segids : numpy.ndarray
        The new atomwise segment indices.
    """
    from ..topology.base import change_squash

    # the original topology
    top = universe._topology

    # detect the atom level attributes (excluding residue level and above)
    atom_attrindices = [
        idx
        for idx, each_attr in enumerate(top.attrs)
        if (Residue not in each_attr.target_classes)
        and (not isinstance(each_attr, Atomindices))
    ]
    attrs = [top.attrs[each_attr] for each_attr in atom_attrindices]

    # create new residues level stuff
    residue_attrindices = [
        idx
        for idx, each_attr in enumerate(top.attrs)
        if (
            (Residue in each_attr.target_classes)
            and (Segment not in each_attr.target_classes)
            and (not isinstance(each_attr, Resids))
            and (not isinstance(each_attr, Resindices))
        )
    ]  # residue level attributes except resids and resindices

    res_criteria = [atomwise_resids, atomwise_segids] + [
        getattr(universe.atoms, top.attrs[each_attr].attrname)
        for each_attr in residue_attrindices
        if top.attrs[each_attr].attrname != "resnums"
    ]

    res_to_squash = [atomwise_resids, atomwise_segids] + [
        getattr(universe.atoms, top.attrs[each_attr].attrname)
        for each_attr in residue_attrindices
    ]

    residx, res_squashed = change_squash(res_criteria, res_to_squash)
    resids = res_squashed[0]
    res_squashed_segids = res_squashed[1]
    n_residues = len(resids)
    # all residue-level attributes except resids
    res_squashed_res_attrs = res_squashed[2:]

    res_attrs = [Resids(resids)] + [
        # use the correspdoning type of the attribute with new sqaushed values
        top.attrs[each_attr].__class__(res_squashed_res_attrs[idx])
        for idx, each_attr in enumerate(residue_attrindices)
    ]
    attrs.extend(res_attrs)

    # create new segment level stuff
    segidx, (segids,) = change_squash(
        (res_squashed_segids,), (res_squashed_segids,)
    )
    n_segments = len(segids)
    attrs.append(Segids(segids))

    # other attributes
    other_attrs = [
        each_attr
        for each_attr in top.attrs
        if (
            (Segment in each_attr.target_classes)
            and (not isinstance(each_attr, Segids))
            and (not isinstance(each_attr, Atomindices))
            and (not isinstance(each_attr, Resindices))
            and (not isinstance(each_attr, Segindices))
        )
    ]

    # create new topology
    top = Topology(
        universe.atoms.n_atoms,
        n_residues,
        n_segments,
        attrs=attrs,
        atom_resindex=residx,
        residue_segindex=segidx,
    )

    # add back other attributes
    if len(other_attrs) > 0:
        for each_otherattr in other_attrs:
            top.add_TopologyAttr(each_otherattr)

    # update the topology in the universe
    universe._topology = top


def _check_file_like(topology):
    if isstream(topology):
        if hasattr(topology, "name"):
            _name = topology.name
        else:
            _name = None
        return NamedStream(topology, _name)
    return topology


def _topology_from_file_like(topology_file, topology_format=None, **kwargs):
    parser = get_parser_for(topology_file, format=topology_format)

    try:
        with parser(topology_file) as p:
            topology = p.parse(**kwargs)
    except (IOError, OSError) as err:
        # There are 2 kinds of errors that might be raised here:
        # one because the file isn't present
        # or the permissions are bad, second when the parser fails
        if err.errno is not None and errno.errorcode[err.errno] in [
            "ENOENT",
            "EACCES",
        ]:
            # Runs if the error is propagated due to no permission / file not found
            raise sys.exc_info()[1] from err
        else:
            # Runs when the parser fails
            raise IOError(
                "Failed to load from the topology file {0}"
                " with parser {1}.\n"
                "Error: {2}".format(topology_file, parser, err)
            )
    except (ValueError, NotImplementedError) as err:
        raise ValueError(
            "Failed to construct topology from file {0}"
            " with parser {1}.\n"
            "Error: {2}".format(topology_file, parser, err)
        )
    return topology


def _resolve_formats(*coordinates, format=None, topology_format=None):
    if not coordinates:
        if format is None:
            format = topology_format
        elif topology_format is None:
            topology_format = format
    return format, topology_format


def _resolve_coordinates(
    filename, *coordinates, format=None, all_coordinates=False
):
    if all_coordinates or not coordinates and filename is not None:
        try:
            get_reader_for(filename, format=format)
        except (ValueError, TypeError):
            warnings.warn(
                "No coordinate reader found for {}. Skipping "
                "this file.".format(filename)
            )
        else:
            coordinates = (filename,) + coordinates
    return coordinates


def _generate_from_topology(universe):
    # generate Universe version of each class
    # AG, RG, SG, A, R, S
    universe._class_bases, universe._classes = groups.make_classes()

    # Put Group level stuff from topology into class
    for attr in universe._topology.attrs:
        universe._process_attr(attr)

    # Generate atoms, residues and segments.
    # These are the first such groups generated for this universe, so
    #  there are no cached merged classes yet. Otherwise those could be
    #  used directly to get a (very) small speedup. (Only really pays off
    #  the readability loss if instantiating millions of AtomGroups at
    #  once.)
    universe.atoms = AtomGroup(np.arange(universe._topology.n_atoms), universe)

    universe.residues = ResidueGroup(
        np.arange(universe._topology.n_residues), universe
    )

    universe.segments = SegmentGroup(
        np.arange(universe._topology.n_segments), universe
    )


class Universe(object):
    """The MDAnalysis Universe contains all the information describing the system.

    The system always requires a *topology file* --- in the simplest case just
    a list of atoms. This can be a CHARMM/NAMD PSF file or a simple coordinate
    file with atom informations such as XYZ, PDB, GROMACS GRO or TPR, or CHARMM
    CRD. See :ref:`Supported topology formats` for what kind of topologies can
    be read.

    A *trajectory file* provides coordinates; the coordinates have to be
    ordered in the same way as the list of atoms in the topology. A trajectory
    can be a single frame such as a PDB, CRD, or GRO file, or it can be a MD
    trajectory (in CHARMM/NAMD/LAMMPS DCD, GROMACS XTC/TRR, AMBER nc, generic
    XYZ format, ...).  See :ref:`Supported coordinate formats` for what can be
    read as a "trajectory".

    As a special case, when the topology is a file that contains atom
    information *and* coordinates (such as XYZ, PDB, GRO or CRD, see
    :ref:`Supported coordinate formats`) then the coordinates are immediately
    loaded from the "topology" file unless a trajectory is supplied.

    Parameters
    ----------
    topology: str, stream, Topology, numpy.ndarray, None
        A CHARMM/XPLOR PSF topology file, PDB file or Gromacs GRO file; used to
        define the list of atoms. If the file includes bond information,
        partial charges, atom masses, ... then these data will be available to
        MDAnalysis. Alternatively, an existing
        :class:`MDAnalysis.core.topology.Topology` instance may be given,
        numpy coordinates, or ``None`` for an empty universe.
    coordinates: str, stream, list of str, list of stream (optional)
        Coordinates can be provided as files of
        a single frame (eg a PDB, CRD, or GRO file); a list of single
        frames; or a trajectory file (in CHARMM/NAMD/LAMMPS DCD, Gromacs
        XTC/TRR, or generic XYZ format). The coordinates must be
        ordered in the same way as the list of atoms in the topology.
        See :ref:`Supported coordinate formats` for what can be read
        as coordinates. Alternatively, streams can be given.
    topology_format: str, ``None``, default ``None``
        Provide the file format of the topology file; ``None`` guesses it from
        the file extension. Can also pass a subclass of
        :class:`MDAnalysis.topology.base.TopologyReaderBase` to define a custom
        reader to be used on the topology file.
    format: str, ``None``, default ``None``
        Provide the file format of the coordinate or trajectory file; ``None``
        guesses it from the file extension. Note that this keyword has no
        effect if a list of file names is supplied because the "chained" reader
        has to guess the file format for each individual list member.
        Can also pass a subclass of :class:`MDAnalysis.coordinates.base.ProtoReader`
        to define a custom reader to be used on the trajectory file.
    all_coordinates: bool, default ``False``
        If set to ``True`` specifies that if more than one filename is passed
        they are all to be used, if possible, as coordinate files (employing a
        :class:`MDAnalysis.coordinates.chain.ChainReader`). The
        default behavior is to take the first file as a topology and the
        remaining as coordinates. The first argument will always always be used
        to infer a topology regardless of *all_coordinates*.
    guess_bonds: bool, default ``False``
        Once Universe has been loaded, attempt to guess the connectivity
        between atoms.  This will populate the .bonds, .angles, and .dihedrals
        attributes of the Universe.

        .. deprecated:: 2.8.0
           This keyword is deprecated and will be removed in MDAnalysis 3.0.
           Please pass ("bonds", "angles", "dihedrals") into
           `to_guess` or `force_guess` instead to guess bonds, angles,
           and dihedrals respectively.

    vdwradii: dict, ``None``, default ``None``
        For use with *guess_bonds*. Supply a dict giving a vdwradii for each
        atom type which are used in guessing bonds.

        .. deprecated:: 2.8.0
           This keyword is deprecated and will be removed in MDAnalysis 3.0.
           Please pass it into Guesser creation (:mod:`~MDAnalysis.guesser`),
           or to :meth:`~MDAnalysis.core.universe.Universe.guess_TopologyAttrs`
           method instead. If passed into `guess_TopologyAttrs`, it will
           override the values set during Guesser creation.

    context: str or :mod:`Guesser<MDAnalysis.guesser>`, default ``'default'``
        Type of the Guesser to be used in guessing TopologyAttrs
    to_guess: list[str] (optional, default ``['types', 'masses']``)
        TopologyAttrs to be guessed. These TopologyAttrs will be wholly
        guessed if they don't exist in the Universe. If they already exist in
        the Universe, only empty or missing values will be guessed.

        .. warning::
            In MDAnalysis 2.x, types and masses are being automatically guessed
            if they are missing (``to_guess=('types, 'masses')``).
            However, starting with release 3.0 **no guessing will be done
            by default** and it will be up to the user to request guessing
            using ``to_guess`` and ``force_guess``.

    force_guess: list[str], (optional)
        TopologyAttrs in this list will be force guessed. If the TopologyAttr
        does not already exist in the Universe, this has no effect. If the
        TopologyAttr does already exist, all values will be overwritten
        by guessed values.
    fudge_factor: float, default [0.55]
        For use with *guess_bonds*. Supply the factor by which atoms must
        overlap each other to be considered a bond.

        .. deprecated:: 2.8.0
           This keyword is deprecated and will be removed in MDAnalysis 3.0.
           Please pass it into Guesser creation (:mod:`~MDAnalysis.guesser`),
           or to :meth:`~MDAnalysis.core.universe.Universe.guess_TopologyAttrs`
           method instead. If passed into `guess_TopologyAttrs`, it will
           override the values set during Guesser creation.

    lower_bound: float, default [0.1]
        For use with *guess_bonds*. Supply the minimum bond length.

        .. deprecated:: 2.8.0
           This keyword is deprecated and will be removed in MDAnalysis 3.0.
           Please pass it into Guesser creation (:mod:`~MDAnalysis.guesser`),
           or to :meth:`~MDAnalysis.core.universe.Universe.guess_TopologyAttrs`
           method instead. If passed into `guess_TopologyAttrs`, it will
           override the values set during Guesser creation.


    transformations: function or list, ``None``, default ``None``
        Provide a list of transformations that you wish to apply to the
        trajectory upon reading. Transformations can be found in
        :mod:`MDAnalysis.transformations`, or can be user-created.
    in_memory: bool, default ``False``
        After reading in the trajectory, transfer it to an in-memory
        representations, which allow for manipulation of coordinates.
    in_memory_step: int, default 1
        Only read every nth frame into in-memory representation.
    continuous: bool, default ``False``
        The `continuous` option is used by the
        :mod:`ChainReader<MDAnalysis.coordinates.chain>`, which contains the
        functionality to treat independent trajectory files as a single virtual
        trajectory.
    **kwargs: extra arguments are passed to the topology parser.
        For instance, when reading a PDB file
        (:class:`PDBReader<MDAnalysis.coordinates.PDB>`,
        :class:`PDBParser<MDAnalysis.topology.PDBParser>`), set
        ``force_chainids_to_segids=True`` to make the universe use the
        chainIDs (column 22) instead of the segmentIDs (column 73-76) as the
        `segids` in the universe and select the corresponding SegmentGroup.

    Attributes
    ----------
    trajectory : base.ReaderBase or base.SingleFrameReaderBase
        currently loaded trajectory reader; readers are described in
        :ref:`Coordinates`
    dimensions : numpy.ndarray
        system dimensions (simulation unit cell, if set in the
        trajectory) at the *current time step*
        (see :attr:`MDAnalysis.coordinates.base.Timestep.dimensions`).
        The unit cell can be set for the current time step (but the change is
        not permanent unless written to a file).
    atoms : AtomGroup
        all particles (:class:`~MDAnalysis.core.groups.Atom`) in the system,
        as read from the `topology` file
    residues : ResidueGroup
        all residues (:class:`~MDAnalysis.core.groups.Residue`) in the system
    segments : SegmentGroup
        all segments (:class:`~MDAnalysis.core.groups.Segment`) in the system
    bonds : topologyattrs.Bonds
        all bonds (if defined in the `topology`) as provided by
        :attr:`Universe.atoms.bonds`
    angles : topologyattrs.Angles
        all angles (if defined in the `topology`), same as
        :attr:`Universe.atoms.angles`
    dihedrals : topologyattrs.Dihedrals
        all dihedral angles (if defined in the `topology`), same as
        :attr:`Universe.atoms.dihedrals`
    impropers : topologyattrs.Impropers
        all improper dihedral angles (if defined in the `topology`), same as
        :attr:`Universe.atoms.impropers`

    Examples
    --------
    Examples for setting up a :class:`Universe`::

       u = Universe(topology, trajectory)          # read system from file(s)
       u = Universe(pdbfile)                       # read atoms and coordinates from PDB or GRO
       u = Universe(topology, [traj1, traj2, ...]) # read from a list of trajectories
       u = Universe(topology, traj1, traj2, ...)   # read from multiple trajectories

    Load new data into a universe (replaces old trajectory and does *not* append)::

       u.load_new(trajectory)                      # read from a new trajectory file

    Selecting atoms with :meth:`~Universe.select_atoms` ::

       ag = u.select_atoms(...)

    returns an :class:`~MDAnalysis.core.groups.AtomGroup`.


    .. versionchanged:: 1.0.0
        Universe() now raises an error. Use Universe(None) or :func:`Universe.empty()` instead.
        Removed instant selectors.

    .. versionchanged:: 2.0.0
        Universe now can be (un)pickled.
        ``topology`` and ``trajectory`` are reserved upon unpickle.
    .. versionchanged:: 2.5.0
        Added fudge_factor and lower_bound parameters for use with
        *guess_bonds*.

    .. versionchanged:: 2.8.0
        Added :meth:`~MDAnalysis.core.universe.Universe.guess_TopologyAttrs` API
        guessing masses and atom types after topology
        is read from a registered parser.

    .. versionchanged:: 2.10.0
        Added :meth: `~MDAnalysis.core.universe.Universe.set_groups`
        API to set residues/segments based on the atomwise resids/segids.

    """

    def __init__(
        self,
        topology=None,
        *coordinates,
        all_coordinates=False,
        format=None,
        topology_format=None,
        transformations=None,
        guess_bonds=False,
        vdwradii=None,
        fudge_factor=0.55,
        lower_bound=0.1,
        in_memory=False,
        context="default",
        to_guess=("types", "masses"),
        force_guess=(),
        in_memory_step=1,
        **kwargs,
    ):

        self._trajectory = None  # managed attribute holding Reader
        self._cache = {"_valid": {}}
        self.atoms = None
        self.residues = None
        self.segments = None
        self.filename = None
        self._context = get_guesser(context)
        self._kwargs = {
            "transformations": transformations,
            "guess_bonds": guess_bonds,
            "vdwradii": vdwradii,
            "fudge_factor": fudge_factor,
            "lower_bound": lower_bound,
            "in_memory": in_memory,
            "in_memory_step": in_memory_step,
            "format": format,
            "topology_format": topology_format,
            "all_coordinates": all_coordinates,
        }
        self._kwargs.update(kwargs)

        format, topology_format = _resolve_formats(
            *coordinates, format=format, topology_format=topology_format
        )

        if not isinstance(topology, Topology) and not topology is None:
            self.filename = _check_file_like(topology)
            topology = _topology_from_file_like(
                self.filename, topology_format=topology_format, **kwargs
            )

        if topology is not None:
            self._topology = topology
        else:
            # point to Universe.empty instead of making empty universe
            raise TypeError(
                "Topology argument required to make Universe. "
                "Try Universe.empty(n_atoms, ...) to construct "
                "your own Universe."
            )

        _generate_from_topology(self)  # make real atoms, res, segments

        coordinates = _resolve_coordinates(
            self.filename,
            *coordinates,
            format=format,
            all_coordinates=all_coordinates,
        )

        if coordinates:
            self.load_new(
                coordinates,
                format=format,
                in_memory=in_memory,
                in_memory_step=in_memory_step,
                **kwargs,
            )

        if transformations:
            if callable(transformations):
                transformations = [transformations]
            self._trajectory.add_transformations(*transformations)

        if guess_bonds:
            warnings.warn(
                "Bond guessing through the `guess_bonds` keyword is deprecated"
                " and will be removed in MDAnalysis 3.0. "
                "Instead, pass 'bonds', 'angles', and 'dihedrals' to "
                "the `to_guess` keyword in Universe for guessing these "
                "if they are not present, or `force_guess` if they are "
                "and you wish to replace these bonds with guessed values. "
                "The kwargs `fudge_factor`, `vdwradii`, and `lower_bound` "
                "are also deprecated and will be removed in MDAnalysis 3.0, "
                "where they should be passed into the Context for guessing on "
                "Universe instantiation. If using guess_TopologyAttrs, "
                "pass these kwargs to the method instead, as they will override "
                "the previous Context values.",
                DeprecationWarning,
            )
            # Original behaviour is to add additionally guessed bond info
            # this is achieved by adding to the `to_guess` list (unliked `force_guess`
            # which replaces existing bonds).
            to_guess = list(to_guess) + ["bonds", "angles", "dihedrals"]

        self.guess_TopologyAttrs(
            context, to_guess, force_guess, error_if_missing=False
        )

    def copy(self):
        """Return an independent copy of this Universe"""
        context = self._context.copy()
        new = self.__class__(
            self._topology.copy(), to_guess=(), context=context
        )
        new.trajectory = self.trajectory.copy()
        return new

    @classmethod
    def empty(
        cls,
        n_atoms,
        n_residues=1,
        n_segments=1,
        n_frames=1,
        atom_resindex=None,
        residue_segindex=None,
        trajectory=False,
        velocities=False,
        forces=False,
    ):
        """Create a blank Universe

        Useful for building a Universe without requiring existing files,
        for example for system building.

        If `trajectory` is set to ``True``, a
        :class:`MDAnalysis.coordinates.memory.MemoryReader` will be
        attached to the Universe.

        Parameters
        ----------
        n_atoms: int
          number of Atoms in the Universe
        n_residues: int, default 1
          number of Residues in the Universe, defaults to 1
        n_segments: int, default 1
          number of Segments in the Universe, defaults to 1
        n_frames: int, default 1
          number of Frames in the Universe, defaults to 1
        atom_resindex: array like, optional
          mapping of atoms to residues, e.g. with 6 atoms,
          `atom_resindex=[0, 0, 1, 1, 2, 2]` would put 2 atoms
          into each of 3 residues.
        residue_segindex: array like, optional
          mapping of residues to segments
        trajectory: bool, optional
          if ``True``, attaches a
          :class:`MDAnalysis.coordinates.memory.MemoryReader` allowing
          coordinates to be set and written.
        velocities: bool, optional
          include velocities in the
          :class:`MDAnalysis.coordinates.memory.MemoryReader`
        forces: bool, optional
          include forces in the
          :class:`MDAnalysis.coordinates.memory.MemoryReader`

        Returns
        -------
        Universe
          :class:`~MDAnalysis.core.universe.Universe` instance with dummy
          values for atoms and undefined coordinates/velocities/forces

        Examples
        --------
        For example to create a new Universe with 6 atoms in 2 residues, with
        positions for the atoms and a mass attribute::

          u = mda.Universe.empty(6, 2,
                                 atom_resindex=np.array([0, 0, 0, 1, 1, 1]),
                                 trajectory=True,
                )
          u.add_TopologyAttr('masses')

        .. versionadded:: 0.17.0
        .. versionchanged:: 0.19.0
           The attached Reader when trajectory=True is now a MemoryReader
        .. versionchanged:: 1.0.0
           Universes can now be created with 0 atoms

        """
        if not n_atoms:
            n_residues = 0
            n_segments = 0

        if atom_resindex is None and n_residues > 1:
            warnings.warn(
                "Residues specified but no atom_resindex given.  "
                "All atoms will be placed in first Residue.",
                UserWarning,
            )

        if residue_segindex is None and n_segments > 1:
            warnings.warn(
                "Segments specified but no segment_resindex given.  "
                "All residues will be placed in first Segment",
                UserWarning,
            )

        top = Topology(
            n_atoms,
            n_residues,
            n_segments,
            atom_resindex=atom_resindex,
            residue_segindex=residue_segindex,
        )

        u = cls(top, to_guess=())

        if n_frames > 1 or trajectory:
            coords = np.zeros((n_frames, n_atoms, 3), dtype=np.float32)
            vels = np.zeros_like(coords) if velocities else None
            forces = np.zeros_like(coords) if forces else None

            # grab and attach a MemoryReader
            u.trajectory = get_reader_for(coords)(
                coords,
                order="fac",
                n_atoms=n_atoms,
                velocities=vels,
                forces=forces,
            )

        return u

    @property
    def universe(self):
        # for Writer.write(universe), see Issue 49
        # Encapsulation in an accessor prevents the Universe from
        # having to keep a reference to itself,
        #  which might be undesirable if it has a __del__ method.
        # It is also cleaner than a weakref.
        return self

    def load_new(
        self,
        filename,
        format=None,
        in_memory=False,
        in_memory_step=1,
        **kwargs,
    ):
        """Load coordinates from `filename`.

        The file format of `filename` is autodetected from the file name suffix
        or can be explicitly set with the `format` keyword. A sequence of files
        can be read as a single virtual trajectory by providing a list of
        filenames.


        Parameters
        ----------
        filename: str or list
            the coordinate file (single frame or trajectory) *or* a list of
            filenames, which are read one after another.
        format: str or list or object (optional)
            provide the file format of the coordinate or trajectory file;
            ``None`` guesses it from the file extension. Note that this
            keyword has no effect if a list of file names is supplied because
            the "chained" reader has to guess the file format for each
            individual list member [``None``]. Can also pass a subclass of
            :class:`MDAnalysis.coordinates.base.ProtoReader` to define a custom
            reader to be used on the trajectory file.
        in_memory: bool (optional)
            Directly load trajectory into memory with the
            :class:`~MDAnalysis.coordinates.memory.MemoryReader`

            .. versionadded:: 0.16.0

        **kwargs: dict
            Other kwargs are passed to the trajectory reader (only for
            advanced use)

        Returns
        -------
        universe: Universe

        Raises
        ------
        TypeError
             if trajectory format can not be determined or no appropriate
             trajectory reader found


        .. versionchanged:: 0.8
           If a list or sequence that is provided for `filename` only contains
           a single entry then it is treated as single coordinate file. This
           has the consequence that it is not read by the
           :class:`~MDAnalysis.coordinates.chain.ChainReader` but directly by
           its specialized file format reader, which typically has more
           features than the
           :class:`~MDAnalysis.coordinates.chain.ChainReader`.

        .. versionchanged:: 0.17.0
           Now returns a :class:`Universe` instead of the tuple of file/array
           and detected file type.
        .. versionchanged:: 2.4.0
           Passes through kwargs if `in_memory=True`.

        """
        # filename==None happens when only a topology is provided
        if filename is None:
            return self

        if len(util.asiterable(filename)) == 1:
            # make sure a single filename is not handed to the ChainReader
            filename = util.asiterable(filename)[0]
        logger.debug("Universe.load_new(): loading {0}...".format(filename))

        try:
            reader = get_reader_for(filename, format=format)
        except ValueError as err:
            raise TypeError(
                "Cannot find an appropriate coordinate reader for file '{0}'.\n"
                "           {1}".format(filename, err)
            )

        # supply number of atoms for readers that cannot do it for themselves
        kwargs["n_atoms"] = self.atoms.n_atoms

        self.trajectory = reader(filename, format=format, **kwargs)
        if self.trajectory.n_atoms != len(self.atoms):
            raise ValueError(
                "The topology and {form} trajectory files don't"
                " have the same number of atoms!\n"
                "Topology number of atoms {top_n_atoms}\n"
                "Trajectory: {fname} Number of atoms {trj_n_atoms}".format(
                    form=self.trajectory.format,
                    top_n_atoms=len(self.atoms),
                    fname=filename,
                    trj_n_atoms=self.trajectory.n_atoms,
                )
            )

        if in_memory:
            self.transfer_to_memory(step=in_memory_step, **kwargs)

        return self

    def transfer_to_memory(
        self, start=None, stop=None, step=None, verbose=False, **kwargs
    ):
        """Transfer the trajectory to in memory representation.

        Replaces the current trajectory reader object with one of type
        :class:`MDAnalysis.coordinates.memory.MemoryReader` to support in-place
        editing of coordinates.

        Parameters
        ----------
        start: int, optional
            start reading from the nth frame.
        stop: int, optional
            read upto and excluding the nth frame.
        step: int, optional
            Read in every nth frame. [1]
        verbose: bool, optional
            Will print the progress of loading trajectory to memory, if
            set to True. Default value is False.


        .. versionadded:: 0.16.0
        .. versionchanged:: 2.4.0
           Passes through kwargs to MemoryReader
        """
        from ..coordinates.memory import MemoryReader

        if not isinstance(self.trajectory, MemoryReader):
            n_frames = len(
                range(*self.trajectory.check_slice_indices(start, stop, step))
            )
            n_atoms = len(self.atoms)
            coordinates = np.zeros((n_frames, n_atoms, 3), dtype=np.float32)
            ts = self.trajectory.ts
            has_vels = ts.has_velocities
            has_fors = ts.has_forces
            has_dims = ts.dimensions is not None

            velocities = np.zeros_like(coordinates) if has_vels else None
            forces = np.zeros_like(coordinates) if has_fors else None
            dimensions = (
                np.zeros((n_frames, 6), dtype=np.float32) if has_dims else None
            )

            for i, ts in enumerate(
                ProgressBar(
                    self.trajectory[start:stop:step],
                    verbose=verbose,
                    desc="Loading frames",
                )
            ):
                np.copyto(coordinates[i], ts.positions)
                if has_vels:
                    np.copyto(velocities[i], ts.velocities)
                if has_fors:
                    np.copyto(forces[i], ts.forces)
                if has_dims:
                    np.copyto(dimensions[i], ts.dimensions)

            # Overwrite trajectory in universe with an MemoryReader
            # object, to provide fast access and allow coordinates
            # to be manipulated
            if step is None:
                step = 1
            self.trajectory = MemoryReader(
                coordinates,
                dimensions=dimensions,
                dt=self.trajectory.ts.dt * step,
                filename=self.trajectory.filename,
                velocities=velocities,
                forces=forces,
                **kwargs,
            )

    # python 2 doesn't allow an efficient splitting of kwargs in function
    # argument signatures.
    # In python3-only we'd be able to explicitly define this function with
    # something like (sel, *othersels, updating=False, **selgroups)
    def select_atoms(self, *args, **kwargs):
        """Select atoms.

        See Also
        --------
        :meth:`MDAnalysis.core.groups.AtomGroup.select_atoms`
        """
        return self.atoms.select_atoms(*args, **kwargs)

    @property
    def bonds(self):
        """Bonds between atoms.

        :meta private:
        """
        return self.atoms.bonds

    @property
    def angles(self):
        """Angles between atoms.

        :meta private:
        """
        return self.atoms.angles

    @property
    def dihedrals(self):
        """Dihedral angles between atoms.

        :meta private:
        """
        return self.atoms.dihedrals

    @property
    def impropers(self):
        """Improper dihedral angles between atoms.

        :meta private:
        """
        return self.atoms.impropers

    def __repr__(self):
        # return "<Universe with {n_atoms} atoms{bonds}>".format(
        #    n_atoms=len(self.atoms),
        #    bonds=" and {0} bonds".format(len(self.bonds)) if self.bonds else "")

        return "<Universe with {n_atoms} atoms>".format(
            n_atoms=len(self.atoms)
        )

    @classmethod
    def _unpickle_U(cls, top, traj, context):
        """Special method used by __reduce__ to deserialise a Universe"""
        #  top is a Topology obj at this point, but Universe can handle that.
        u = cls(top, to_guess=(), context=context)
        u.trajectory = traj

        return u

    def __reduce__(self):
        #  __setstate__/__getstate__ will raise an error when Universe has a
        #  transformation (that has AtomGroup inside). Use __reduce__ instead.
        #  Universe's two "legs" of top and traj both serialise themselves.
        return (
            self._unpickle_U,
            (self._topology, self._trajectory, self._context.copy()),
        )

    # Properties
    @property
    def dimensions(self):
        """Current dimensions of the unitcell.

        :meta private:
        """
        return self.coord.dimensions

    @dimensions.setter
    def dimensions(self, box):
        """Set dimensions if the Timestep allows this

        .. versionadded:: 0.9.0
        """
        # Add fancy error handling here or use Timestep?
        self.coord.dimensions = box

    @property
    def coord(self):
        """Reference to current timestep and coordinates of universe.

        The raw trajectory coordinates are :attr:`Universe.coord.positions`,
        represented as a :class:`numpy.float32` array.

        Because :attr:`coord` is a reference to a
        :class:`~MDAnalysis.coordinates.timestep.Timestep`, it changes its contents
        while one is stepping through the trajectory.

        .. Note::

           In order to access the coordinates it is better to use the
           :meth:`AtomGroup.positions` method; for instance, all coordinates of
           the Universe as a numpy array: :meth:`Universe.atoms.positions`.

        """
        return self.trajectory.ts

    @property
    def kwargs(self):
        """keyword arguments used to initialize this universe"""
        return copy.deepcopy(self._kwargs)

    @property
    def trajectory(self):
        """Reference to trajectory reader object containing trajectory data.

        :meta private:
        """
        if self._trajectory is not None:
            return self._trajectory
        else:
            raise AttributeError("No trajectory loaded into Universe")

    @trajectory.setter
    def trajectory(self, value):
        del self._trajectory  # guarantees that files are closed (?)
        self._trajectory = value

    def add_TopologyAttr(self, topologyattr, values=None):
        """Add a new topology attribute to the Universe

        Adding a TopologyAttribute to the Universe makes it available to
        all AtomGroups etc throughout the Universe.

        Parameters
        ----------
        topologyattr: TopologyAttr or string
          Either a MDAnalysis TopologyAttr object or the name of a possible
          topology attribute.
        values: np.ndarray, optional
          If initiating an attribute from a string, the initial values to
          use.  If not supplied, the new TopologyAttribute will have empty
          or zero values.

        Example
        -------
        For example to add bfactors to a Universe:

        >>> import MDAnalysis as mda
        >>> from MDAnalysis.tests.datafiles import PSF, DCD
        >>> u = mda.Universe(PSF, DCD)
        >>> u.add_TopologyAttr('tempfactors')
        >>> u.atoms.tempfactors
        array([0., 0., 0., ..., 0., 0., 0.])

        .. versionchanged:: 0.17.0
           Can now also add TopologyAttrs with a string of the name of the
           attribute to add (eg 'charges'), can also supply initial values
           using values keyword.

        .. versionchanged:: 1.1.0
            Now warns when adding bfactors to a Universe with
            existing tempfactors, or adding tempfactors to a
            Universe with existing bfactors.
            In version 2.0, MDAnalysis will stop treating
            tempfactors and bfactors as separate attributes. Instead,
            they will be aliases of the same attribute.
        """
        if isinstance(topologyattr, str):
            if topologyattr in ("bfactor", "bfactors"):
                warnings.warn(BFACTOR_WARNING, DeprecationWarning)
            try:
                tcls = _TOPOLOGY_ATTRS[topologyattr]
            except KeyError:
                errmsg = (
                    "Unrecognised topology attribute name: '{}'."
                    "  Possible values: '{}'\n"
                    "To raise an issue go to: "
                    "https://github.com/MDAnalysis/mdanalysis/issues"
                    "".format(
                        topologyattr, ", ".join(sorted(_TOPOLOGY_ATTRS.keys()))
                    )
                )
                raise ValueError(errmsg) from None
            else:
                topologyattr = tcls.from_blank(
                    n_atoms=self._topology.n_atoms,
                    n_residues=self._topology.n_residues,
                    n_segments=self._topology.n_segments,
                    values=values,
                )
        self._topology.add_TopologyAttr(topologyattr)
        self._process_attr(topologyattr)

    def del_TopologyAttr(self, topologyattr):
        """Remove a topology attribute from the Universe

        Removing a TopologyAttribute from the Universe makes it unavailable to
        all AtomGroups etc throughout the Universe.

        Parameters
        ----------
        topologyattr: TopologyAttr or string
          Either a MDAnalysis TopologyAttr object or the name of a possible
          topology attribute.

        Example
        -------
        For example to remove bfactors to a Universe:

        >>> import MDAnalysis as mda
        >>> from MDAnalysis.tests.datafiles import PSF, DCD
        >>> u = mda.Universe(PSF, DCD)
        >>> u.add_TopologyAttr('tempfactors')
        >>> hasattr(u.atoms[:3], 'tempfactors')
        True
        >>>
        >>> u.del_TopologyAttr('tempfactors')
        >>> hasattr(u.atoms[:3], 'tempfactors')
        False


        .. versionadded:: 2.0.0
        """

        if not isinstance(topologyattr, str):
            try:
                topologyattr = topologyattr.attrname
            except AttributeError:
                # either TopologyGroup or not valid
                try:
                    # this may not end well
                    # e.g. matrix -> matrices
                    topologyattr = topologyattr.btype + "s"
                except AttributeError:
                    raise ValueError(
                        "Topology attribute must be str or "
                        "TopologyAttr object or class. "
                        f"Given: {type(topologyattr)}"
                    ) from None

        try:
            topologyattr = _TOPOLOGY_ATTRS[topologyattr].attrname
        except KeyError:
            attrs = ", ".join(sorted(_TOPOLOGY_ATTRS))
            errmsg = (
                f"Unrecognised topology attribute: '{topologyattr}'."
                f"  Possible values: '{attrs}'\n"
                "To raise an issue go to: "
                "https://github.com/MDAnalysis/mdanalysis/issues"
            )
            raise ValueError(errmsg) from None

        try:
            topattr = getattr(self._topology, topologyattr)
        except AttributeError:
            raise ValueError(
                f"Topology attribute {topologyattr} " "not in Universe."
            ) from None
        self._topology.del_TopologyAttr(topattr)
        self._unprocess_attr(topattr)

    def _process_attr(self, attr):
        """Squeeze a topologyattr for its information

        Grabs:
         - Group properties (attribute access)
         - Component properties
         - Transplant methods
        """
        n_dict = {
            "atom": self._topology.n_atoms,
            "residue": self._topology.n_residues,
            "segment": self._topology.n_segments,
        }
        logger.debug("_process_attr: Adding {0} to topology".format(attr))
        if (
            attr.per_object is not None
            and len(attr) != n_dict[attr.per_object]
        ):
            raise ValueError(
                "Length of {attr} does not"
                " match number of {obj}s.\n"
                "Expect: {n:d} Have: {m:d}".format(
                    attr=attr.attrname,
                    obj=attr.per_object,
                    n=n_dict[attr.per_object],
                    m=len(attr),
                )
            )

        for cls in attr.target_classes:
            self._class_bases[cls]._add_prop(attr)

        # TODO: Try and shove this into cls._add_prop
        # Group transplants
        for cls in (
            Atom,
            Residue,
            Segment,
            GroupBase,
            AtomGroup,
            ResidueGroup,
            SegmentGroup,
        ):
            for funcname, meth in attr.transplants[cls]:
                setattr(self._class_bases[cls], funcname, meth)
        # Universe transplants
        for funcname, meth in attr.transplants["Universe"]:
            setattr(self.__class__, funcname, meth)

    def _unprocess_attr(self, attr):
        """
        Undo all the stuff in _process_attr.

        If the topology attribute is not present, nothing happens
        (silent fail).
        """
        for cls in attr.target_classes:
            self._class_bases[cls]._del_prop(attr)

        # Universe transplants
        for funcname, _ in attr.transplants.pop("Universe", []):
            delattr(self.__class__, funcname)
        # Group transplants
        for cls, transplants in attr.transplants.items():
            for funcname, _ in transplants:
                delattr(self._class_bases[cls], funcname)

    def add_Residue(self, segment=None, **attrs):
        """Add a new Residue to this Universe

        New Residues will not contain any Atoms, but can be assigned to Atoms
        as per usual.  If the Universe contains multiple segments, this must
        be specified as a keyword.

        Parameters
        ----------
        segment: MDAnalysis.Segment
          If there are multiple segments, then the Segment that the new
          Residue will belong in must be specified.
        attrs: dict
          For each Residue attribute, the value for the new Residue must be
          specified

        Returns
        -------
        A reference to the new Residue

        Raises
        ------
        NoDataError
          If any information was missing.  This happens before any changes have
          been made, ie the change is rolled back.


        Example
        -------

        Adding a new GLY residue, then placing atoms within it:

        >>> import MDAnalysis as mda
        >>> from MDAnalysis.tests.datafiles import PSF, DCD
        >>> u = mda.Universe(PSF, DCD)
        >>> newres = u.add_Residue(segment=u.segments[0], resid=42, resname='GLY', resnum=0)
        >>> u.atoms[[1, 2, 3]].residues = newres
        >>> u.select_atoms('resname GLY and resid 42 and resnum 0')
        <AtomGroup with 3 atoms>

        """
        if len(self.segments) == 1:  # if only one segment, use this
            segment = self.segments[0]
        if segment is None:
            raise NoDataError("")
        # pass this information to the topology
        residx = self._topology.add_Residue(segment, **attrs)
        # resize my residues
        self.residues = ResidueGroup(
            np.arange(self._topology.n_residues), self
        )

        # return the new residue
        return self.residues[residx]

    def add_Segment(self, **attrs):
        """Add a new Segment to this Universe

        Parameters
        ----------
        attrs: dict
            For each Segment attribute as a key, give the value in the new
            Segment

        Returns
        -------
        A reference to the new Segment

        Raises
        ------
        NoDataError
            If any attributes were not specified as a keyword.

        """
        # pass this information to the topology
        segidx = self._topology.add_Segment(**attrs)
        # resize my segments
        self.segments = SegmentGroup(
            np.arange(self._topology.n_segments), self
        )
        # return the new segment
        return self.segments[segidx]

    def _add_topology_objects(
        self, object_type, values, types=None, guessed=False, order=None
    ):
        """Add new TopologyObjects to this Universe

        Parameters
        ----------
        object_type : {'bonds', 'angles', 'dihedrals', 'impropers'}
            The type of TopologyObject to add.
        values : TopologyGroup or iterable of tuples, AtomGroups, or TopologyObjects
            An iterable of: tuples of atom indices, or AtomGroups,
            or TopologyObjects. If every value is a TopologyObject, all
            keywords are ignored.
            If AtomGroups or TopologyObjects are passed, they *must* be from the same
            Universe.
        types : iterable (optional, default None)
            None, or an iterable of hashable values with the same length as ``values``
        guessed : bool or iterable (optional, default False)
            bool, or an iterable of hashable values with the same length as ``values``
        order : iterable (optional, default None)
            None, or an iterable of hashable values with the same length as ``values``


        .. versionadded:: 1.0.0
        """
        if all(isinstance(x, TopologyObject) for x in values):
            try:
                types = [t.type for t in values]
            except AttributeError:
                types = None
            guessed = [t.is_guessed for t in values]
            order = [t.order for t in values]

        indices = []
        for x in values:
            if isinstance(x, (AtomGroup, TopologyObject)):
                if x.universe is not self:
                    err_msg = "Cannot add {} from different Universes."
                    raise ValueError(err_msg.format(object_type))
                indices.append(x.indices)
            else:
                indices.append(x)

        all_indices = set([i for obj in indices for i in obj])
        nonexistent = all_indices - set(self.atoms.indices)
        if nonexistent:
            istr = ", ".join(map(str, nonexistent))
            err_msg = "Cannot add {} for nonexistent atom indices: {}"
            raise ValueError(err_msg.format(object_type, istr))

        try:
            attr = getattr(self._topology, object_type)
        except AttributeError:
            self.add_TopologyAttr(object_type, [])
            attr = getattr(self._topology, object_type)

        attr._add_bonds(indices, types=types, guessed=guessed, order=order)

    def add_bonds(self, values, types=None, guessed=False, order=None):
        """Add new Bonds to this Universe.

        Parameters
        ----------
        values : iterable of tuples, AtomGroups, or Bonds; or TopologyGroup
            An iterable of: tuples of 2 atom indices, or AtomGroups with 2 atoms,
            or Bonds. If every value is a Bond, all
            keywords are ignored.
            If AtomGroups, Bonds, or a TopologyGroup are passed,
            they *must* be from the same Universe.
        types : iterable (optional, default None)
            None, or an iterable of hashable values with the same length as ``values``
        guessed : bool or iterable (optional, default False)
            bool, or an iterable of hashable values with the same length as ``values``
        order : iterable (optional, default None)
            None, or an iterable of hashable values with the same length as ``values``


        Example
        -------

        Adding TIP4P water bonds with a list of AtomGroups::

            import MDAnalysis as mda
            from MDAnalysis.tests.datafiles import GRO
            u = mda.Universe(GRO)
            sol = u.select_atoms('resname SOL')
            ow_hw1 = sol.select_atoms('name OW or name HW1').split('residue')
            ow_hw2 = sol.select_atoms('name OW or name HW2').split('residue')
            ow_mw = sol.select_atoms('name OW or name MW').split('residue')
            u.add_bonds(ow_hw1 + ow_hw2 + ow_mw)

        You can only add bonds from the same Universe. If you would like to add
        AtomGroups, Bonds, or a TopologyGroup from a different Universe, convert
        them to indices first. ::

            from MDAnalysis.tests.datafiles import PSF
            u2 = mda.Universe(PSF)

            #  assuming you have already added bonds to u
            u2.add_bonds(u.bonds.to_indices())


        .. versionadded:: 1.0.0
        """
        self._add_topology_objects(
            "bonds", values, types=types, guessed=guessed, order=order
        )
        self._invalidate_bond_related_caches()

    def _invalidate_bond_related_caches(self):
        """
        Invalidate caches related to bonds and fragments.

        This should be called whenever the Universe's bonds are modified.

        .. versionadded: 2.8.0
        """
        # Invalidate bond-related caches
        self._cache.pop("fragments", None)
        self._cache["_valid"].pop("fragments", None)
        self._cache["_valid"].pop("fragindices", None)

    def add_angles(self, values, types=None, guessed=False):
        """Add new Angles to this Universe.

        Parameters
        ----------
        values : iterable of tuples, AtomGroups, or Angles; or TopologyGroup
            An iterable of: tuples of 3 atom indices, or AtomGroups with 3 atoms,
            or Angles. If every value is a Angle, all
            keywords are ignored.
            If AtomGroups, Angles, or a TopologyGroup are passed,
            they *must* be from the same Universe.
        types : iterable (optional, default None)
            None, or an iterable of hashable values with the same length as ``values``
        guessed : bool or iterable (optional, default False)
            bool, or an iterable of hashable values with the same length as ``values``

        .. versionadded:: 1.0.0
        """
        self._add_topology_objects(
            "angles", values, types=types, guessed=guessed
        )

    def add_dihedrals(self, values, types=None, guessed=False):
        """Add new Dihedrals to this Universe.

        Parameters
        ----------
        values : iterable of tuples, AtomGroups, or Dihedrals; or TopologyGroup
            An iterable of: tuples of 4 atom indices, or AtomGroups with 4 atoms,
            or Dihedrals. If every value is a Dihedral, all
            keywords are ignored.
            If AtomGroups, Dihedrals, or a TopologyGroup are passed,
            they *must* be from the same Universe.
        types : iterable (optional, default None)
            None, or an iterable of hashable values with the same length as ``values``
        guessed : bool or iterable (optional, default False)
            bool, or an iterable of hashable values with the same length as ``values``


        .. versionadded:: 1.0.0
        """
        self._add_topology_objects(
            "dihedrals", values, types=types, guessed=guessed
        )

    def add_impropers(self, values, types=None, guessed=False):
        """Add new Impropers to this Universe.

        Parameters
        ----------
        values : iterable of tuples, AtomGroups, or Impropers; or TopologyGroup
            An iterable of: tuples of 4 atom indices, or AtomGroups with 4 atoms,
            or Impropers. If every value is an Improper, all
            keywords are ignored.
            If AtomGroups, Impropers, or a TopologyGroup are passed,
            they *must* be from the same Universe.
        types : iterable (optional, default None)
            None, or an iterable of hashable values with the same length as ``values``
        guessed : bool or iterable (optional, default False)
            bool, or an iterable of hashable values with the same length as ``values``


        .. versionadded:: 1.0.0
        """
        self._add_topology_objects(
            "impropers", values, types=types, guessed=guessed
        )

    def _delete_topology_objects(self, object_type, values):
        """Delete TopologyObjects from this Universe

        Parameters
        ----------
        object_type : {'bonds', 'angles', 'dihedrals', 'impropers'}
            The type of TopologyObject to delete.
        values : iterable of tuples, AtomGroups, or TopologyObjects; or TopologyGroup
            An iterable of: tuples of atom indices, or AtomGroups,
            or TopologyObjects.
            If AtomGroups, TopologyObjects, or a TopologyGroup are passed,
            they *must* be from the same Universe.

        .. versionadded:: 1.0.0
        """
        indices = []
        for x in values:
            if isinstance(x, (AtomGroup, TopologyObject)):
                if x.universe is not self:
                    err_msg = "Cannot delete {} from different Universes."
                    raise ValueError(err_msg.format(object_type))
                indices.append(x.indices)
            else:
                indices.append(x)

        try:
            attr = getattr(self._topology, object_type)
        except AttributeError:
            raise ValueError("There are no {} to delete".format(object_type))
        attr._delete_bonds(indices)

    def delete_bonds(self, values):
        """Delete Bonds from this Universe.

        Parameters
        ----------
        values : iterable of tuples, AtomGroups, or Bonds; or TopologyGroup
            An iterable of: tuples of 2 atom indices, or AtomGroups with 2 atoms,
            or Bonds.
            If AtomGroups, Bonds, or a TopologyGroup are passed,
            they *must* be from the same Universe.


        Example
        -------

        Deleting bonds from a Universe::

            import MDAnalysis as mda
            from MDAnalysis.tests.datafiles import PSF
            u = mda.Universe(PSF)

            #  delete first 5 bonds
            u.delete_bonds(u.bonds[:5])


        If you are deleting bonds in the form of AtomGroups, Bonds, or a
        TopologyGroup, they must come from the same Universe. If you want to
        delete bonds from another Universe, convert them to indices first. ::

            from MDAnalysis.tests.datafiles import PDB
            u2 = mda.Universe(PDB)

            u.delete_bonds(u2.bonds.to_indices())


        .. versionadded:: 1.0.0
        """
        self._delete_topology_objects("bonds", values)
        self._invalidate_bond_related_caches()

    def delete_angles(self, values):
        """Delete Angles from this Universe.

        Parameters
        ----------
        values : iterable of tuples, AtomGroups, or Angles; or TopologyGroup
            An iterable of: tuples of 3 atom indices, or AtomGroups with 3 atoms,
            or Angles.
            If AtomGroups, Angles, or a TopologyGroup are passed,
            they *must* be from the same Universe.


        .. versionadded:: 1.0.0
        """
        self._delete_topology_objects("angles", values)

    def delete_dihedrals(self, values):
        """Delete Dihedrals from this Universe.

        Parameters
        ----------
        values : iterable of tuples, AtomGroups, or Dihedrals; or TopologyGroup
            An iterable of: tuples of 4 atom indices, or AtomGroups with 4 atoms,
            or Dihedrals.
            If AtomGroups, Dihedrals, or a TopologyGroup are passed,
            they *must* be from the same Universe.


        .. versionadded:: 1.0.0
        """
        self._delete_topology_objects("dihedrals", values)

    def delete_impropers(self, values):
        """Delete Impropers from this Universe.

        Parameters
        ----------
        values : iterable of tuples, AtomGroups, or Impropers; or TopologyGroup
            An iterable of: tuples of 4 atom indices, or AtomGroups with 4 atoms,
            or Impropers.
            If AtomGroups, Angles, or a TopologyGroup are passed,
            they *must* be from the same Universe.


        .. versionadded:: 1.0.0
        """
        self._delete_topology_objects("impropers", values)

    # TODO: Maybe put this as a Bond attribute transplant
    # Problems: Can we transplant onto Universe?
    # Probably a smarter way to do this too, could generate
    # these on demand *per atom*.
    # Wouldn't then need the Universe linkage here
    #
    # Alternate idea: Bonds Attribute generates a Fragments
    # Attribute (ie, 2 for the price of 1)
    # Fragments then gets its own Class/namespace/jazz.
    @property
    @cached("fragments")
    def _fragdict(self):
        """
        .. versionadded:: 0.9.0
        .. versionchanged:: 0.16.0
           Fragment atoms are sorted by their index, and framgents are sorted
           by their first atom index so their order is predictable.
        .. versionchanged:: 0.19.0
           Uses faster C++ implementation
        .. versionchanged:: 0.20.0
           * _fragdict keys are now atom indices instead of Atoms
           * _fragdict items are now a namedtuple ``fraginfo(ix, fragment)``
             storing the fragindex ``ix`` along with the fragment.
        """
        atoms = self.atoms.ix
        bonds = self.atoms.bonds.to_indices()

        frag_indices = find_fragments(atoms, bonds)
        frags = tuple([AtomGroup(np.sort(ix), self) for ix in frag_indices])

        fragdict = {}
        fraginfo = collections.namedtuple("fraginfo", "ix, fragment")
        for i, f in enumerate(frags):
            for a in f:
                fragdict[a.ix] = fraginfo(i, f)

        return fragdict

    @classmethod
    def from_smiles(
        cls,
        smiles,
        sanitize=True,
        addHs=True,
        generate_coordinates=True,
        numConfs=1,
        rdkit_kwargs=None,
        **kwargs,
    ):
        """Create a Universe from a SMILES string with rdkit

        Parameters
        ----------
        smiles : str
            SMILES string
        sanitize : bool (optional, default True)
            Toggle the sanitization of the molecule
        addHs : bool (optional, default True)
            Add all necessary hydrogens to the molecule
        generate_coordinates : bool (optional, default True)
            Generate 3D coordinates using RDKit's
            :func:`AllChem.EmbedMultipleConfs` function. Requires adding
            hydrogens with the `addHs` parameter
        numConfs : int (optional, default 1)
            Number of frames to generate coordinates for. Ignored if
            ``generate_coordinates=False``
        rdkit_kwargs : dict (optional)
            Other arguments passed to the RDKit :func:`EmbedMultipleConfs`
            function
        kwargs : dict
            Parameters passed on Universe creation

        Returns
        -------
        universe : Universe
            contains atom names and topology information (bonds) derived from
            the input SMILES string; coordinates are included if
            `generate_coordinates` was set to ``True``


        Examples
        --------
        To create a Universe with 10 conformers of ethanol:

        >>> from rdkit.Chem import AllChem
        >>> u = mda.Universe.from_smiles('CCO', numConfs=10)
        >>> u
        <Universe with 9 atoms>
        >>> u.trajectory
        <RDKitReader with 10 frames of 9 atoms>

        To use a different conformer generation algorithm, like ETKDGv3:

        >>> u = mda.Universe.from_smiles('CCO', rdkit_kwargs=dict(
        ...      params=AllChem.ETKDGv3()))
        >>> u.trajectory
        <RDKitReader with 1 frames of 9 atoms>


        .. versionadded:: 2.0.0

        """
        if rdkit_kwargs is None:
            rdkit_kwargs = {}
        try:
            from rdkit import Chem
            from rdkit.Chem import AllChem
        except ImportError as e:
            raise ImportError(
                "Creating a Universe from a SMILES string requires RDKit but "
                "it does not appear to be installed"
            ) from e

        mol = Chem.MolFromSmiles(smiles, sanitize=sanitize)
        if mol is None:
            raise SyntaxError("Error while parsing SMILES {0}".format(smiles))
        if addHs:
            mol = Chem.AddHs(mol)
        if generate_coordinates:
            if not addHs:
                raise ValueError(
                    "Generating coordinates requires adding "
                    "hydrogens with `addHs=True`"
                )

            numConfs = rdkit_kwargs.pop("numConfs", numConfs)
            if not (isinstance(numConfs, int) and numConfs > 0):
                raise SyntaxError(
                    "numConfs must be a non-zero positive "
                    "integer instead of {0}".format(numConfs)
                )
            AllChem.EmbedMultipleConfs(mol, numConfs, **rdkit_kwargs)

        return cls(mol, **kwargs)

    def guess_TopologyAttrs(
        self,
        context=None,
        to_guess=None,
        force_guess=None,
        error_if_missing=True,
        **kwargs,
    ):
        """
        Guess and add attributes through a specific context-aware guesser.

        Parameters
        ----------
        context: str or :mod:`Guesser<MDAanalysis.guesser>` class
            For calling a matching guesser class for this specific context
        to_guess: Optional[list[str]]
            TopologyAttrs to be guessed. These TopologyAttrs will be wholly
            guessed if they don't exist in the Universe. If they already exist in
            the Universe, only empty or missing values will be guessed.

            .. warning::
                In MDAnalysis 2.x, types and masses are being automatically guessed
                if they are missing (``to_guess=('types, 'masses')``).
                However, starting with release 3.0 **no guessing will be done
                by default** and it will be up to the user to request guessing
                using ``to_guess`` and ``force_guess``.

        force_guess: Optional[list[str]]
            TopologyAttrs in this list will be force guessed. If the
            TopologyAttr does not already exist in the Universe, this has no
            effect. If the TopologyAttr does already exist, all values will
            be overwritten by guessed values.
        error_if_missing: bool
            If `True`, raise an error if the guesser cannot guess the attribute
            due to missing TopologyAttrs used as the inputs for guessing.
            If `False`, a warning will be raised instead.
            Errors will always be raised if an attribute is in the
            ``force_guess`` list, even if this parameter is set to False.

        **kwargs: extra arguments to be passed to the guesser class

        Examples
        --------
        To guess ``masses`` and ``types`` attributes::

            u.guess_TopologyAttrs(context='default', to_guess=['masses', 'types'])

        .. versionadded:: 2.8.0

        """
        if not context:
            context = self._context

        # update iteratively to avoid multiple kwargs clashing
        guesser_kwargs = {}
        guesser_kwargs.update(self._kwargs)
        guesser_kwargs.update(kwargs)
        guesser = get_guesser(context, self.universe, **guesser_kwargs)
        self._context = guesser

        if to_guess is None:
            to_guess = []
        if force_guess is None:
            force_guess = []

        total_guess = list(to_guess) + list(force_guess)

        # Removing duplicates from the guess list while keeping attributes
        # order as it is more convenient to guess attributes
        # in the same order that the user provided
        total_guess = list(dict.fromkeys(total_guess))

        # Set of all Connectivity related attribute names
        # used to special case attribute replacement after calling the guesser
        objects = set(
            topattr.attrname
            for topattr in _TOPOLOGY_ATTRS.values()
            if issubclass(topattr, _Connection)
        )

        # Checking if the universe is empty to avoid errors
        # from guesser methods
        if self._topology.n_atoms > 0:

            topology_attrs = [
                att.attrname for att in self._topology.read_attributes
            ]

            common_attrs = set(to_guess) & set(topology_attrs)
            common_attrs = ", ".join(attr for attr in common_attrs)

            if len(common_attrs) > 0:
                logger.info(
                    f"The attribute(s) {common_attrs} have already been read "
                    "from the topology file. The guesser will "
                    "only guess empty values for this attribute, "
                    "if any exists. To overwrite it by completely "
                    "guessed values, you can pass the attribute to"
                    " the force_guess parameter instead of "
                    "the to_guess one"
                )

            for attr in total_guess:
                if guesser.is_guessable(attr):
                    fg = attr in force_guess
                    try:
                        values = guesser.guess_attr(attr, fg)
                    except NoDataError as e:
                        if error_if_missing or fg:
                            raise e
                        else:
                            warnings.warn(str(e))
                            continue

                    # None indicates no additional guessing was done
                    if values is None:
                        continue
                    if attr in objects:
                        # delete existing connections if they exist
                        if fg and hasattr(self.atoms, attr):
                            group = getattr(self.atoms, attr)
                            self._delete_topology_objects(attr, group)
                        # this method appends any new bonds in values to existing bonds
                        self._add_topology_objects(attr, values, guessed=True)
                        if attr == "bonds":
                            self._invalidate_bond_related_caches()
                    else:
                        guessed_attr = _TOPOLOGY_ATTRS[attr](values, True)
                        self.add_TopologyAttr(guessed_attr)
                    logger.info(
                        f"attribute {attr} has been guessed" " successfully."
                    )

                else:
                    raise ValueError(
                        f"{context} guesser can not guess the"
                        f" following attribute: {attr}"
                    )

        else:
            warnings.warn(
                "Can not guess attributes " "for universe with 0 atoms"
            )

    def set_groups(self, atomwise_resids=None, atomwise_segids=None):
        """Set the groups (`ResidueGroup`, `SegmentGroup`) of the Universe
        by atomwise resids/segids.

        The `topology` will also be updated based on the provided `atomwise_resids`
        and `atomwise_segids`. The original `resids` and `segids` will be stored
        in attributes `atomwise_resids_orig` and/or `atomwise_segids_orig` if
        they are modified.
        See notes for the logic of the function.

        Parameters
        ----------
        atomwise_resids:
            A list of residue IDs to be set for the Universe. The length
            of the list should be equal to the number of atoms in the Universe.
            If `None`, the original resids will be used.

        atomwise_segids:
            A list of segment IDs to be set for the Universe. The length
            of the list should be equal to the number of atoms in the Universe.
            If `None`, the original segids will be used.

        Raises
        ------
        AssertionError
            If the length of the provided atomwise_resids or atomwise_segids
            does not match the number of atoms in the Universe.

        Notes
        -----
        First, the function will check if resids or segids is provided.
        If both resids and segids are not provided (`None`), it will do nothing.
        If only one of them is provided, it will use the original values for the
        other one. If both are provided, it will use the provided values for
        both resids and segids.
        The function will then update the topology by a new generated topology
        with new values of the resids and segids.
        Finally, the corresponding new `ResidueGroup` and `SegmentGroup` will be
        created by the updated topology.

        Examples
        --------
        To set custom segment IDs for the segments of the Universe::

            atomwise_segids = ['A', 'A', 'B', 'B']
            u.set_groups(atomwise_segids=atomwise_segids)

            # Now the Universe has two segments with segIDs 'A' and 'B'
            u.segments
            >>> <SegmentGroup with 2 segments>

        .. versionadded:: 2.10.0
        """
        if (atomwise_resids is None) and (atomwise_segids is None):
            warnings.warn(
                "Not setting groups. Please provide atomwise_resids or "
                "atomwise_segids."
            )
            return

        # resids
        if atomwise_resids is None:
            atomwise_resids = self.atoms.resids

        else:
            # check the length of atomwise_resids
            if len(atomwise_resids) != self.atoms.n_atoms:
                raise ValueError(
                    "The length of atomwise_resids should be the same as "
                    "the number of atoms in the universe."
                )

            self.atomwise_resids_orig = self.atoms.resids
            logger.info(
                "The new resids replaces the current one. "
                "The original resids is stored in "
                "atomwise_resids_orig."
            )

        # segids
        if atomwise_segids is None:
            atomwise_segids = self.atoms.segids

        else:
            # check the length of atomwise_segids
            if len(atomwise_segids) != self.atoms.n_atoms:
                raise ValueError(
                    "The length of atomwise_segids should be the same as "
                    "the number of atoms in the universe."
                )

            self.atomwise_segids_orig = self.atoms.segids
            logger.info(
                "The new resids replaces the current one. "
                "The original segids is stored in "
                "atomwise_segids_orig."
            )

        atomwise_resids = np.array(atomwise_resids, dtype=int)
        atomwise_segids = np.array(atomwise_segids, dtype=object)

        _update_topology_by_ids(
            self,
            atomwise_resids=atomwise_resids,
            atomwise_segids=atomwise_segids,
        )
        _generate_from_topology(self)


def Merge(*args):
    """Create a new new :class:`Universe` from one or more
    :class:`~MDAnalysis.core.groups.AtomGroup` instances.

    Parameters
    ----------
    *args: :class:`~MDAnalysis.core.groups.AtomGroup`
        One or more AtomGroups.

    Returns
    -------
    universe: :class:`Universe`

    Raises
    ------
    ValueError
        Too few arguments or an AtomGroup is empty and
    TypeError
        Arguments are not :class:`AtomGroup` instances.

    Notes
    -----
    The resulting :class:`Universe` will only inherit the common topology
    attributes that all merged universes share.

    :class:`AtomGroup` instances can come from different Universes, or can come
    directly from a :meth:`~Universe.select_atoms` call.

    :class:`Merge` can also be used with a single :class:`AtomGroup` if the
    user wants to, for example, re-order the atoms in the :class:`Universe`.

    If multiple :class:`AtomGroup` instances from the same :class:`Universe`
    are given, the merge will first simply "add" together the
    :class:`AtomGroup` instances.

    Merging does not create a full trajectory but only a single structure even
    if the input consists of one or more trajectories.  However, one can use
    the :class:`~MDAnalysis.coordinates.memory.MemoryReader` to construct a
    trajectory for the new Universe as described under
    :ref:`creating-in-memory-trajectory-label`.

    Example
    -------
    In this example, protein, ligand, and solvent were externally prepared in
    three different PDB files. They are loaded into separate :class:`Universe`
    objects (where they could be further manipulated, e.g. renumbered,
    relabeled, rotated, ...) The :func:`Merge` command is used to combine all
    of them together::

       u1 = Universe("protein.pdb")
       u2 = Universe("ligand.pdb")
       u3 = Universe("solvent.pdb")
       u = Merge(u1.select_atoms("protein"), u2.atoms, u3.atoms)
       u.atoms.write("system.pdb")

    The complete system is then written out to a new PDB file.


    .. versionchanged:: 0.9.0
       Raises exceptions instead of assertion errors.

    .. versionchanged:: 0.16.0
       The trajectory is now a
       :class:`~MDAnalysis.coordinates.memory.MemoryReader`.

    """
    from ..topology.base import squash_by

    if len(args) == 0:
        raise ValueError("Need at least one AtomGroup for merging")

    for a in args:
        if not isinstance(a, groups.AtomGroup):
            raise TypeError(repr(a) + " is not an AtomGroup")
    for a in args:
        if len(a) == 0:
            raise ValueError("cannot merge empty AtomGroup")

    # Create a new topology using the intersection of topology attributes
    blank_topology_attrs = set(dir(Topology(attrs=[])))
    common_attrs = set.intersection(
        *[set(dir(ag.universe._topology)) for ag in args]
    )
    tops = set(["bonds", "angles", "dihedrals", "impropers"])

    attrs = []

    # Create set of attributes which are array-valued and can be simply
    # concatenated together
    common_array_attrs = common_attrs - blank_topology_attrs - tops
    # Build up array-valued topology attributes including only attributes
    # that all arguments' universes have
    for attrname in common_array_attrs:
        for ag in args:
            attr = getattr(ag, attrname)
            attr_class = type(getattr(ag.universe._topology, attrname))
            if issubclass(attr_class, AtomAttr):
                pass
            elif issubclass(attr_class, ResidueAttr):
                attr = getattr(ag.residues, attrname)
            elif issubclass(attr_class, SegmentAttr):
                attr = getattr(ag.segments, attrname)
            else:
                raise NotImplementedError(
                    "Don't know how to handle"
                    " TopologyAttr not subclassed"
                    " from AtomAttr, ResidueAttr,"
                    " or SegmentAttr."
                )
            if type(attr) != np.ndarray:
                raise TypeError(
                    "Encountered unexpected topology "
                    "attribute of type {}".format(type(attr))
                )
            try:
                attr_array.extend(attr)
            except NameError:
                attr_array = list(attr)
        attrs.append(attr_class(np.array(attr_array, dtype=attr.dtype)))
        del attr_array

    # Build up topology groups including only those that all arguments'
    # universes have
    for t in tops & common_attrs:
        offset = 0
        bondidx = []
        types = []
        for ag in args:
            # create a mapping scheme for this atomgroup
            mapping = {a.index: i for i, a in enumerate(ag, start=offset)}
            offset += len(ag)

            tg = getattr(ag, t)
            bonds_class = type(getattr(ag.universe._topology, t))
            # Create a topology group of only bonds that are within this ag
            # ie we don't want bonds that extend out of the atomgroup
            tg = tg.atomgroup_intersection(ag, strict=True)

            # Map them so they refer to our new indices
            new_idx = [
                tuple([mapping[x] for x in entry]) for entry in tg.indices
            ]
            bondidx.extend(new_idx)
            if hasattr(tg, "_bondtypes"):
                types.extend(tg._bondtypes)
            else:
                types.extend([None] * len(tg))
        if any(t is None for t in types):
            attrs.append(bonds_class(bondidx))
        else:
            types = np.array(types, dtype="|S8")
            attrs.append(bonds_class(bondidx, types))

    # Renumber residue and segment indices
    n_atoms = sum([len(ag) for ag in args])
    residx = []
    segidx = []
    res_offset = 0
    seg_offset = 0
    for ag in args:
        # create a mapping scheme for this atomgroup's parents
        res_mapping = {
            r.resindex: i for i, r in enumerate(ag.residues, start=res_offset)
        }
        seg_mapping = {
            r.segindex: i for i, r in enumerate(ag.segments, start=seg_offset)
        }
        res_offset += len(ag.residues)
        seg_offset += len(ag.segments)

        # Map them so they refer to our new indices
        residx.extend([res_mapping[x] for x in ag.resindices])
        segidx.extend([seg_mapping[x] for x in ag.segindices])

    residx = np.array(residx, dtype=np.int32)
    segidx = np.array(segidx, dtype=np.int32)

    _, _, [segidx] = squash_by(residx, segidx)

    n_residues = len(set(residx))
    n_segments = len(set(segidx))

    top = Topology(
        n_atoms,
        n_residues,
        n_segments,
        attrs=attrs,
        atom_resindex=residx,
        residue_segindex=segidx,
    )

    # Create and populate a universe
    try:
        # Create universe with coordinates if they exists in args
        coords = np.vstack([a.positions for a in args])
        u = Universe(
            top,
            coords[None, :, :],
            format=MDAnalysis.coordinates.memory.MemoryReader,
        )
    except AttributeError:
        # Create universe without coordinates if they dont exists in args
        u = Universe(top)

    return u



"""Tests for core.groups.requires decorator

"""

import numpy as np

import pytest
from MDAnalysis.core.groups import requires
from MDAnalysis import NoDataError

from MDAnalysisTests import make_Universe


class TestRequires(object):
    def test_requires_failure_singular(self):
        @requires("masses")
        def mass_multiplier(ag1, ag2, scalar):
            return (ag1.masses + ag2.masses) * scalar

        u = make_Universe(("charges",))
        with pytest.raises(NoDataError):
            mass_multiplier(u.atoms[:10], u.atoms[20:30], 4.0)

    def test_requires_failure_multiple(self):
        @requires("masses", "charges")
        def mass_multiplier(ag1, ag2, scalar):
            return (ag1.masses + ag2.charges) * scalar

        u = make_Universe(("masses", "types"))
        with pytest.raises(NoDataError):
            mass_multiplier(u.atoms[:10], u.atoms[20:30], 4.0)

    def test_requires_success(self):
        @requires("masses")
        def mass_multiplier(ag1, ag2, scalar):
            return (ag1.masses + ag2.masses) * scalar

        u = make_Universe(("masses",))

        result = mass_multiplier(u.atoms[:10], u.atoms[20:30], 4.0)

        assert isinstance(result, np.ndarray)

    def test_failure_errormessage(self):
        # failures should list all required attributes not
        # just the first one
        @requires("cats", "dogs", "frogs")
        def animal_print(ag):
            return len(ag.cats), len(ag.dogs), len(ag.frogs)

        u = make_Universe()
        try:
            animal_print(u.atoms)
        except NoDataError as e:
            message = e.args[0]
            # Test function name gets returned (for debug)
            assert "animal_print" in message
            assert "cats" in message
            assert "dogs" in message
            assert "frogs" in message
        else:
            pytest.fail(msg="Should raise NoDataError")



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787

r"""
Default Guesser
================
.. _DefaultGuesser:

DefaultGuesser is a generic guesser class that has basic guessing methods.
This class is a general purpose guesser that can be used with most topologies,
but being generic makes it the least accurate among all guessers.


Guessing behavior
-----------------

This section describes how each attribute is guessed by the DefaultGuesser.

Masses
~~~~~~

We first attempt to look up the mass of an atom based on its element if the
element TopologyAttr is available. If not, we attempt to lookup the mass based
on the atom type (``type``) TopologyAttr. If neither of these is available, we
attempt to guess the atom type based on the atom name (``name``) and then
lookup the mass based on the guessed atom type.


Types
~~~~~

We attempt to guess the atom type based on the atom name (``name``).
The name is first stripped of any numbers and symbols, and then looked up in
the :data:`MDAnalysis.guesser.tables.atomelements` table. If the name is not
found, we continue checking variations of the name following the logic in
:meth:`DefaultGuesser.guess_atom_element`. Ultimately, if no match is found,
the first character of the stripped name is returned.

Elements
~~~~~~~~

This follows the same method as guessing atom types.


Bonds
~~~~~

Bonds are guessed based on the distance between atoms.
See :meth:`DefaultGuesser.guess_bonds` for more details.

Angles
~~~~~~

Angles are guessed based on the bonds between atoms.
See :meth:`DefaultGuesser.guess_angles` for more details.

Dihedrals
~~~~~~~~~

Dihedrals are guessed based on the angles between atoms.
See :meth:`DefaultGuesser.guess_dihedrals` for more details.

Improper Dihedrals
~~~~~~~~~~~~~~~~~~

Improper dihedrals are guessed based on the angles between atoms.
See :meth:`DefaultGuesser.guess_improper_dihedrals` for more details.

Aromaticities
~~~~~~~~~~~~~

Aromaticity is guessed using RDKit's GetIsAromatic method.
See :meth:`DefaultGuesser.guess_aromaticities` for more details.




Classes
-------

.. autoclass:: DefaultGuesser
   :members:
   :inherited-members:

"""
from .base import GuesserBase
import numpy as np
import warnings
import math

import re

from ..exceptions import NoDataError
from ..lib import distances
from . import tables


class DefaultGuesser(GuesserBase):
    """
    This guesser holds generic methods (not directed to specific contexts) for
    guessing different topology attribute. It has the same methods which where
    originally found in Topology.guesser.py. The attributes that can be
    guessed by this class are:
     - masses
     - types
     - elements
     - angles
     - dihedrals
     - bonds
     - improper dihedrals
     - aromaticities

    You can use this guesser either directly through an instance, or through
    the :meth:`~MDAnalysis.core.universe.Universe.guess_TopologyAttrs` method.

    Parameters
    ----------
    universe : Universe
        The Universe to apply the guesser on
    box : np.ndarray, optional
        The box of the Universe. This is used for bond guessing.
    vdwradii : dict, optional
        Dict relating atom types: vdw radii. This is used for bond guessing
    fudge_factor : float, optional
        The factor by which atoms must overlap each other to be considered
        a bond.  Larger values will increase the number of bonds found. [0.55]
    lower_bound : float, optional
        The minimum bond length. All bonds found shorter than this length
        will be ignored. This is useful for parsing PDB with altloc records
        where atoms with altloc A and B may be very close together and
        there should be no chemical bond between them. [0.1]

    Examples
    --------
    to guess bonds for a universe::

        import MDAnalysis as mda
        from MDAnalysisTests.datafiles import two_water_gro

        u = mda.Universe(two_water_gro, context='default', to_guess=['bonds'])

    .. versionadded:: 2.8.0

    """

    context = "default"

    def __init__(
        self,
        universe,
        box=None,
        vdwradii=None,
        fudge_factor=0.55,
        lower_bound=0.1,
        **kwargs,
    ):
        super().__init__(
            universe,
            box=box,
            vdwradii=vdwradii,
            fudge_factor=fudge_factor,
            lower_bound=lower_bound,
            **kwargs,
        )
        self._guesser_methods = {
            "masses": self.guess_masses,
            "types": self.guess_types,
            "elements": self.guess_types,
            "bonds": self.guess_bonds,
            "angles": self.guess_angles,
            "dihedrals": self.guess_dihedrals,
            "impropers": self.guess_improper_dihedrals,
            "aromaticities": self.guess_aromaticities,
        }

    def guess_masses(self, atom_types=None, indices_to_guess=None):
        """Guess the mass of many atoms based upon their type.
        For guessing masses through :meth:`~MDAnalysis.core.universe.Universe.guess_TopologyAttrs`:

        First try to guess masses from atom elements, if not available,
        try to guess masses from types and if not available, try to guess
        types.

        Parameters
        ----------
        atom_types : Optional[np.ndarray]
          Atom types/elements to guess masses from
        indices_to_guess : Optional[np.ndarray]
          Mask array for partially guess masses for certain atoms

        Returns
        -------
        atom_masses : np.ndarray dtype float64

        Raises
        ------
        :exc:`ValueError`
            If there are no atom types or elements to guess mass from.

        """
        if atom_types is None:
            try:
                atom_types = self._universe.atoms.elements
            except NoDataError:
                try:
                    atom_types = self._universe.atoms.types
                except NoDataError:
                    try:
                        atom_types = self.guess_types(
                            atom_types=self._universe.atoms.names
                        )
                    except NoDataError:
                        raise NoDataError(
                            "there is no reference attributes"
                            " (elements, types, or names)"
                            " in this universe to guess mass from"
                        ) from None

        if indices_to_guess is not None:
            atom_types = atom_types[indices_to_guess]

        masses = np.array(
            [self.get_atom_mass(atom) for atom in atom_types], dtype=np.float64
        )
        return masses

    def get_atom_mass(self, element):
        """Return the atomic mass in u for *element*.
        Masses are looked up in :data:`MDAnalysis.guesser.tables.masses`.

        .. Warning:: Until version 3.0.0 unknown masses are set to 0.0

        """
        try:
            return tables.masses[element]
        except KeyError:
            try:
                return tables.masses[element.upper()]
            except KeyError:
                warnings.warn(
                    "Unknown masses are set to 0.0 for current version, "
                    "this will be deprecated in version 3.0.0 and replaced by"
                    " Masse's no_value_label (np.nan)",
                    PendingDeprecationWarning,
                )
                return 0.0

    def guess_atom_mass(self, atomname):
        """Guess a mass based on the atom name.

        :func:`guess_atom_element` is used to determine the kind of atom.

        .. warning:: Until version 3.0.0 anything not recognized is simply
           set to 0.0; if you rely on the masses you might want to double-check.
        """
        return self.get_atom_mass(self.guess_atom_element(atomname))

    def guess_types(self, atom_types=None, indices_to_guess=None):
        """Guess the atom type of many atoms based on atom name

        Parameters
        ----------
        atom_types (optional)
          atoms names if types guessing is desired to be from names
        indices_to_guess (optional)
          Mask array for partially guess types for certain atoms

        Returns
        -------
        atom_types : np.ndarray dtype object

        Raises
        ------
        :exc:`ValueError`
           If there is no names to guess types from.

        """
        if atom_types is None:
            try:
                atom_types = self._universe.atoms.names
            except NoDataError:
                raise NoDataError(
                    "there is no reference attributes in this universe "
                    "to guess types from"
                ) from None

        if indices_to_guess is not None:
            atom_types = atom_types[indices_to_guess]

        return np.array(
            [self.guess_atom_element(atom) for atom in atom_types],
            dtype=object,
        )

    def guess_atom_element(self, atomname):
        """Guess the element of the atom from the name.

        First all numbers and symbols are stripped from the name.
        Then the name is looked up in the
        :data:`MDAnalysis.guesser.tables.atomelements` table.
        If the name is not found, we remove the last character or
        first character from the name and check the table for both,
        with a preference for removing the last character. If the name is
        still not found, we iteratively continue to remove the last character
        or first character until we find a match. If ultimately no match
        is found, the first character of the stripped name is returned.

        If the input name is an empty string, an empty string is returned.

        The table comes from CHARMM and AMBER atom
        types, where the first character is not sufficient to determine the
        atom type. Some GROMOS ions have also been added.

        .. Warning: The translation table is incomplete.
           This will probably result in some mistakes,
           but it still better than nothing!

        See Also
        --------
        :func:`guess_atom_type`
        :mod:`MDAnalysis.guesser.tables`
        """
        NUMBERS = re.compile(r"[0-9]")  # match numbers
        SYMBOLS = re.compile(r"[*+-]")  # match *, +, -
        if atomname == "":
            return ""
        try:
            return tables.atomelements[atomname.upper()]
        except KeyError:
            # strip symbols and numbers
            no_symbols = re.sub(SYMBOLS, "", atomname)
            name = re.sub(NUMBERS, "", no_symbols).upper()

            # just in case
            if name in tables.atomelements:
                return tables.atomelements[name]

            while name:
                if name in tables.elements:
                    return name
                if name[:-1] in tables.elements:
                    return name[:-1]
                if name[1:] in tables.elements:
                    return name[1:]
                if len(name) <= 2:
                    return name[0]
                name = name[:-1]  # probably element is on left not right

            # if it's numbers
            return no_symbols

    def guess_bonds(self, atoms=None, coords=None):
        r"""Guess if bonds exist between two atoms based on their distance.

        Bond between two atoms is created, if the two atoms are within

        .. math::

              d < f \cdot (R_1 + R_2)

        of each other, where :math:`R_1` and :math:`R_2` are the VdW radii
        of the atoms and :math:`f` is an ad-hoc *fudge_factor*. This is
        the `same algorithm that VMD uses`_.

        Parameters
        ----------
        atoms: AtomGroup
            atoms for which bonds should be guessed
        coords: np.ndarray, optional
            coordinates of the atoms. If not provided, the coordinates
            of the ``atoms`` in the universe are used.

        Returns
        -------
        list
            List of tuples suitable for use in Universe topology building.

        Warnings
        --------
        No check is done after the bonds are guessed to see if Lewis
        structure is correct. This is wrong and will burn somebody.

        Raises
        ------
        :exc:`ValueError`
           If inputs are malformed or `vdwradii` data is missing.


        .. _`same algorithm that VMD uses`:
           http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.1/ug/node26.html

        """
        if atoms is None:
            atoms = self._universe.atoms

        if coords is None:
            coords = self._universe.atoms.positions

        if len(atoms) != len(coords):
            raise ValueError("'atoms' and 'coord' must be the same length")

        fudge_factor = self._kwargs.get("fudge_factor", 0.55)

        # so I don't permanently change it
        vdwradii = tables.vdwradii.copy()
        user_vdwradii = self._kwargs.get("vdwradii", None)
        # this should make algo use their values over defaults
        if user_vdwradii:
            vdwradii.update(user_vdwradii)

        # Try using types, then elements
        if hasattr(atoms, "types"):
            atomtypes = atoms.types
        else:
            atomtypes = self.guess_types(atom_types=atoms.names)

        # check that all types have a defined vdw
        if not all(val in vdwradii for val in set(atomtypes)):
            raise ValueError(
                (
                    "vdw radii for types: "
                    + ", ".join(
                        [t for t in set(atomtypes) if t not in vdwradii]
                    )
                    + ". These can be defined manually using the"
                    + f" keyword 'vdwradii'"
                )
            )

        lower_bound = self._kwargs.get("lower_bound", 0.1)

        box = self._kwargs.get("box", None)

        if box is not None:
            box = np.asarray(box)

        # to speed up checking, calculate what the largest possible bond
        # atom that would warrant attention.
        # then use this to quickly mask distance results later
        max_vdw = max([vdwradii[t] for t in atomtypes])

        bonds = []

        pairs, dist = distances.self_capped_distance(
            coords, max_cutoff=2.0 * max_vdw, min_cutoff=lower_bound, box=box
        )
        for idx, (i, j) in enumerate(pairs):
            d = (
                vdwradii[atomtypes[i]] + vdwradii[atomtypes[j]]
            ) * fudge_factor
            if dist[idx] < d:
                bonds.append((atoms[i].index, atoms[j].index))
        return tuple(bonds)

    def guess_angles(self, bonds=None):
        """Given a list of Bonds, find all angles that exist between atoms.

        Works by assuming that if atoms 1 & 2 are bonded, and 2 & 3 are bonded,
        then (1,2,3) must be an angle.

        Parameters
        ----------
        bonds : Bonds
             from which angles should be guessed

        Returns
        -------
        list of tuples
            List of tuples defining the angles.
            Suitable for use in u._topology


        See Also
        --------
        :meth:`guess_bonds`

        """
        from ..core.universe import Universe

        angles_found = set()

        if bonds is None:
            if hasattr(self._universe.atoms, "bonds"):
                bonds = self._universe.atoms.bonds
            else:
                temp_u = Universe.empty(n_atoms=len(self._universe.atoms))
                temp_u.add_bonds(
                    self.guess_bonds(
                        self._universe.atoms, self._universe.atoms.positions
                    )
                )
                bonds = temp_u.atoms.bonds

        for b in bonds:
            for atom in b:
                other_a = b.partner(atom)  # who's my friend currently in Bond
                for other_b in atom.bonds:
                    if other_b != b:  # if not the same bond I start as
                        third_a = other_b.partner(atom)
                        desc = tuple(
                            [other_a.index, atom.index, third_a.index]
                        )
                        # first index always less than last
                        if desc[0] > desc[-1]:
                            desc = desc[::-1]
                        angles_found.add(desc)

        return tuple(angles_found)

    def guess_dihedrals(self, angles=None):
        """Given a list of Angles, find all dihedrals that exist between atoms.

        Works by assuming that if (1,2,3) is an angle, and 3 & 4 are bonded,
        then (1,2,3,4) must be a dihedral.

        Parameters
        ----------
        angles : Angles
             from which dihedrals should be guessed

        Returns
        -------
        list of tuples
            List of tuples defining the dihedrals.
            Suitable for use in u._topology

        """
        from ..core.universe import Universe

        if angles is None:
            if hasattr(self._universe.atoms, "angles"):
                angles = self._universe.atoms.angles

            else:
                temp_u = Universe.empty(n_atoms=len(self._universe.atoms))

                temp_u.add_bonds(
                    self.guess_bonds(
                        self._universe.atoms, self._universe.atoms.positions
                    )
                )

                temp_u.add_angles(self.guess_angles(temp_u.atoms.bonds))

                angles = temp_u.atoms.angles

        dihedrals_found = set()

        for b in angles:
            a_tup = tuple([a.index for a in b])  # angle as tuple of numbers
            # if searching with b[0], want tuple of (b[2], b[1], b[0], +new)
            # search the first and last atom of each angle
            for atom, prefix in zip(
                [b.atoms[0], b.atoms[-1]], [a_tup[::-1], a_tup]
            ):
                for other_b in atom.bonds:
                    if not other_b.partner(atom) in b:
                        third_a = other_b.partner(atom)
                        desc = prefix + (third_a.index,)
                        if desc[0] > desc[-1]:
                            desc = desc[::-1]
                        dihedrals_found.add(desc)

        return tuple(dihedrals_found)

    def guess_improper_dihedrals(self, angles=None):
        """Given a list of Angles, find all improper dihedrals
        that exist between atoms.

        Works by assuming that if (1,2,3) is an angle, and 2 & 4 are bonded,
        then (2, 1, 3, 4) must be an improper dihedral.
        ie the improper dihedral is the angle between the planes formed by
        (1, 2, 3) and (1, 3, 4)

        Returns
        -------
            List of tuples defining the improper dihedrals.
            Suitable for use in u._topology

        """

        from ..core.universe import Universe

        if angles is None:
            if hasattr(self._universe.atoms, "angles"):
                angles = self._universe.atoms.angles

            else:
                temp_u = Universe.empty(n_atoms=len(self._universe.atoms))

                temp_u.add_bonds(
                    self.guess_bonds(
                        self._universe.atoms, self._universe.atoms.positions
                    )
                )

                temp_u.add_angles(self.guess_angles(temp_u.atoms.bonds))

                angles = temp_u.atoms.angles

        dihedrals_found = set()

        for b in angles:
            atom = b[1]  # select middle atom in angle
            # start of improper tuple
            a_tup = tuple([b[a].index for a in [1, 2, 0]])
            # if searching with b[1], want tuple of (b[1], b[2], b[0], +new)
            # search the first and last atom of each angle
            for other_b in atom.bonds:
                other_atom = other_b.partner(atom)
                # if this atom isn't in the angle I started with
                if other_atom not in b:
                    desc = a_tup + (other_atom.index,)
                    if desc[0] > desc[-1]:
                        desc = desc[::-1]
                    dihedrals_found.add(desc)

        return tuple(dihedrals_found)

    def guess_atom_charge(self, atoms):
        """Guess atom charge from the name.

        .. Warning:: Not implemented; simply returns 0.
        """
        # TODO: do something slightly smarter, at least use name/element
        return 0.0

    def guess_aromaticities(self, atomgroup=None):
        """Guess aromaticity of atoms using RDKit

        Returns
        -------
        aromaticities : numpy.ndarray
            Array of boolean values for the aromaticity of each atom

        """
        if atomgroup is None:
            atomgroup = self._universe.atoms

        mol = atomgroup.convert_to("RDKIT")
        return np.array([atom.GetIsAromatic() for atom in mol.GetAtoms()])

    def guess_gasteiger_charges(self, atomgroup):
        """Guess Gasteiger partial charges using RDKit

        Parameters
        ----------
        atomgroup : mda.core.groups.AtomGroup
            Atoms for which the charges will be guessed

        Returns
        -------
        charges : numpy.ndarray
            Array of float values representing the charge of each atom

        """

        mol = atomgroup.convert_to("RDKIT")
        from rdkit.Chem.rdPartialCharges import ComputeGasteigerCharges

        ComputeGasteigerCharges(mol, throwOnParamFailure=True)
        return np.array(
            [
                atom.GetDoubleProp("_GasteigerCharge")
                for atom in mol.GetAtoms()
            ],
            dtype=np.float32,
        )
